"""
This module contains the SFTPWriter class and the SFTPWriteMode enum.

The SFTPWriter class is used to write data to a file on an SFTP server.
It uses the Paramiko library to establish an SFTP connection and write data to the server.
The data to be written is provided by a BufferWriter, which generates the data in a buffer.
See the docstring of the SFTPWriter class for more details.
Refer to koheesio.spark.writers.buffer for more details on the BufferWriter interface.

The SFTPWriteMode enum defines the different write modes that the SFTPWriter can use.
These modes determine how the SFTPWriter behaves when the file it is trying to write to already exists on the server.
For more details on each mode, see the docstring of the SFTPWriteMode enum.
"""

from typing import Optional, Union
from enum import Enum
import hashlib
from pathlib import Path
import time

from paramiko.sftp_client import SFTPClient
from paramiko.transport import Transport

from pydantic import PrivateAttr

from koheesio.models import (
    Field,
    InstanceOf,
    SecretStr,
    field_validator,
    model_validator,
)
from koheesio.spark.writers import Writer
from koheesio.spark.writers.buffer import (
    BufferWriter,
    PandasCsvBufferWriter,
    PandasJsonBufferWriter,
)

__all__ = ["SFTPWriteMode", "SFTPWriter", "SendCsvToSftp", "SendJsonToSftp"]


# pylint: disable=E1101
class SFTPWriteMode(str, Enum):
    """
    The different write modes for the SFTPWriter.

    ### OVERWRITE:
    * If the file exists, it will be overwritten.
    * If it does not exist, a new file will be created.

    ### APPEND:
    * If the file exists, the new data will be appended to it.
    * If it does not exist, a new file will be created.

    ### IGNORE:
    * If the file exists, the method will return without writing anything.
    * If it does not exist, a new file will be created.

    ### EXCLUSIVE:
    * If the file exists, an error will be raised.
    * If it does not exist, a new file will be created.

    ### BACKUP:
    * If the file exists and the new data is different from the existing data, a backup will be created and the file
        will be overwritten.
    * If it does not exist, a new file will be created.

    ### UPDATE:
    * If the file exists and the new data is different from the existing data, the file will be overwritten.
    * If the file exists and the new data is the same as the existing data, the method will return without writing
        anything.
    * If the file does not exist, a new file will be created.
    """

    OVERWRITE = "overwrite"
    APPEND = "append"
    IGNORE = "ignore"
    EXCLUSIVE = "exclusive"
    BACKUP = "backup"
    UPDATE = "update"

    @classmethod
    def from_string(cls, mode: str) -> "SFTPWriteMode":
        """Return the SFTPWriteMode for the given string."""
        return cls[mode.upper()]

    @property
    def write_mode(self) -> str:
        """Return the write mode for the given SFTPWriteMode."""
        if self in {SFTPWriteMode.OVERWRITE, SFTPWriteMode.BACKUP, SFTPWriteMode.EXCLUSIVE, SFTPWriteMode.UPDATE}:
            return "wb"  # Overwrite, Backup, Exclusive, Update modes set the file to be written from the beginning
        if self == SFTPWriteMode.APPEND:
            return "ab"  # Append mode sets the file to be written from the end


class SFTPWriter(Writer):
    """
    Write a Dataframe to SFTP through a BufferWriter

    Concept
    -------
    * This class uses Paramiko to connect to an SFTP server and write the contents of a buffer to a file on the server.
    * This implementation takes inspiration from https://github.com/springml/spark-sftp

    Parameters
    ----------
    path : Union[str, Path]
        Path to the folder to write to
    file_name : Optional[str], optional, default=None
        Name of the file. If not provided, the file name is expected to be part of the path. Make sure to add the
        desired file extension.
    host : str
        SFTP Host
    port : int
        SFTP Port
    username : SecretStr, optional, default=None
        SFTP Server Username
    password : SecretStr, optional, default=None
        SFTP Server Password
    buffer_writer : BufferWriter
        This is the writer that will generate the body of the file that will be written to the specified
        file through SFTP. Details on how the DataFrame is written to the buffer should be implemented in the
        implementation of the BufferWriter class. Any BufferWriter can be used here, as long as it implements the
        BufferWriter interface.
    mode : SFTPWriteMode, optional, default=SFTPWriteMode.OVERWRITE
        Write mode: overwrite, append, ignore, exclusive, backup, or update. See the docstring of SFTPWriteMode for
        more details.
    """

    path: Union[str, Path] = Field(default=..., description="Path to the folder to write to", alias="prefix")
    file_name: Optional[str] = Field(
        default=None,
        description="Name of the file. If not provided, the file name is expected to be part of the path. Make sure to "
        "add the desired file extension!",
        alias="filename",
    )
    host: str = Field(default=..., description="SFTP Host")
    port: int = Field(default=..., description="SFTP Port")
    username: Optional[SecretStr] = Field(default=None, description="SFTP Server Username")
    password: Optional[SecretStr] = Field(default=None, description="SFTP Server Password")

    buffer_writer: InstanceOf[BufferWriter] = Field(
        default=...,
        description="This is the writer that will generate the body of the file that will be written to the specified "
        "file through SFTP. Details on how the DataFrame is written to the buffer should be implemented in the "
        "implementation of the BufferWriter class. Any BufferWriter can be used here, as long as it implements the "
        "BufferWriter interface.",
    )

    mode: SFTPWriteMode = Field(
        default=SFTPWriteMode.OVERWRITE,
        description="Write mode: overwrite, append, ignore, exclusive, backup, or update." + SFTPWriteMode.__doc__,  # type: ignore
    )

    # private attrs
    _client: Optional[SFTPClient] = PrivateAttr(default=None)
    _transport: Optional[Transport] = PrivateAttr(default=None)

    @model_validator(mode="before")
    def validate_path_and_file_name(cls, data: dict) -> dict:
        """Validate the path, make sure path and file_name are Path objects."""
        path_or_str = data.get("path")

        if isinstance(path_or_str, str):
            # make sure the path is a Path object
            path_or_str = Path(path_or_str)

        if not isinstance(path_or_str, Path):
            raise ValueError(f"Invalid path: {path_or_str}")

        if file_name := data.get("file_name", data.get("filename")):
            path_or_str = path_or_str / file_name
            try:
                del data["filename"]
            except KeyError:
                pass
            data["file_name"] = file_name

        data["path"] = path_or_str
        return data

    @field_validator("host")
    def validate_sftp_host(cls, host: str) -> str:
        """Validate the host"""
        # remove the sftp:// prefix if present
        if host.startswith("sftp://"):
            host = host.replace("sftp://", "")

        # remove the trailing slash if present
        if host.endswith("/"):
            host = host[:-1]

        return host

    @property
    def write_mode(self) -> str:
        """Return the write mode for the given SFTPWriteMode."""
        mode = SFTPWriteMode.from_string(self.mode)  # Convert string to SFTPWriteMode
        return mode.write_mode

    @property
    def transport(self) -> Transport:
        """Return the transport for the SFTP connection. If it doesn't exist, create it.

        If the username and password are provided, use them to connect to the SFTP server.
        """
        if not self._transport:
            self._transport = Transport((self.host, self.port))
            if self.username and self.password:
                self._transport.connect(
                    username=self.username.get_secret_value(), password=self.password.get_secret_value()
                )
            else:
                self._transport.connect()
        return self._transport

    @property
    def client(self) -> SFTPClient:
        """Return the SFTP client. If it doesn't exist, create it."""
        if not self._client:
            try:
                self._client = SFTPClient.from_transport(self.transport)
            except EOFError as e:
                self.log.error(f"Failed to create SFTP client. Transport active: {self.transport.is_active()}")
                raise e
        return self._client

    def _close_client(self) -> None:
        """Close the SFTP client and transport."""
        if self.client:
            self.client.close()
        if self.transport:
            self.transport.close()

    def write_file(self, file_path: str, buffer_output: InstanceOf[BufferWriter.Output]) -> None:
        """
        Using Paramiko, write the data in the buffer to SFTP.
        """
        with self.client.open(file_path, self.write_mode) as file:
            self.log.debug(f"Writing file {file_path} to SFTP...")
            file.write(buffer_output.read())

    def check_file_exists(self, file_path: str) -> bool:
        """
        Check if a file exists on the SFTP server.
        """
        try:
            self.client.stat(file_path)
            return True
        except IOError:
            return False

    def _handle_write_mode(self, file_path: str, buffer_output: InstanceOf[BufferWriter.Output]) -> None:
        """
        Handle different write modes.

        See SFTPWriteMode for more details.
        """
        if not self.check_file_exists(file_path) or self.mode in {SFTPWriteMode.OVERWRITE, SFTPWriteMode.APPEND}:
            # If the file doesn't exist, write the file (irrespective of the mode)
            # Overwrite and Append modes will write the file irrespective of whether it exists or not
            self.write_file(file_path, buffer_output)
            return

        if self.mode == SFTPWriteMode.EXCLUSIVE:
            # If the file exists in EXCLUSIVE mode, raise an error
            raise FileExistsError(f"File {file_path} already exists on the SFTP server.")

        if self.mode == SFTPWriteMode.IGNORE:
            # If the file exists in IGNORE mode, return without writing
            return

        # If the file exists and the mode is UPDATE or BACKUP, check if the new data is different from the existing data
        new_data = buffer_output.read()
        sha256_new_data = hashlib.sha256(new_data).hexdigest()

        with self.client.open(file_path, "rb") as file:
            existing_data = file.read()
            sha256_existing_data = hashlib.sha256(existing_data).hexdigest()

        if sha256_existing_data == sha256_new_data:
            # If the new data is the same as the existing data, return without writing
            return

        # If the new data is different from the existing data
        if self.mode == SFTPWriteMode.BACKUP:
            # If the mode is BACKUP, create a backup
            timestamp = int(time.time())
            backup_file_path = f"{file_path}.{timestamp}.bak"
            self.log.debug(f"Creating backup of {file_path} as {backup_file_path}...")
            self.client.rename(file_path, backup_file_path)

        # Then overwrite the file
        self.write_file(file_path, buffer_output)

    def execute(self) -> Writer.Output:
        buffer_output: InstanceOf[BufferWriter.Output] = self.buffer_writer.write(self.df)

        # write buffer to the SFTP server
        try:
            self._handle_write_mode(self.path.as_posix(), buffer_output)
        finally:
            self._close_client()


class SendCsvToSftp(PandasCsvBufferWriter, SFTPWriter):
    """
    Write a DataFrame to an SFTP server as a CSV file.

    This class uses the PandasCsvBufferWriter to generate the CSV data and the SFTPWriter to write the data to the
    SFTP server.

    Example
    -------
    ```python
    from koheesio.spark.writers import SendCsvToSftp

    writer = SendCsvToSftp(
        # SFTP Parameters
        host="sftp.example.com",
        port=22,
        username="user",
        password="password",
        path="/path/to/folder",
        file_name="file.tsv.gz",
        # CSV Parameters
        header=True,
        sep="\t",
        quote='"',
        timestampFormat="%Y-%m-%d",
        lineSep=os.linesep,
        compression="gzip",
        index=False,
    )

    writer.write(df)
    ```

    In this example, the DataFrame `df` is written to the file `file.csv.gz` in the folder `/path/to/folder` on the
    SFTP server. The file is written as a CSV file with a tab delimiter (TSV), double quotes as the quote character,
    and gzip compression.

    Parameters
    ----------
    path : Union[str, Path]
        Path to the folder to write to.
    file_name : Optional[str]
        Name of the file. If not provided, it's expected to be part of the path.
    host : str
        SFTP Host.
    port : int
        SFTP Port.
    username : SecretStr
        SFTP Server Username.
    password : SecretStr
        SFTP Server Password.
    mode : SFTPWriteMode
        Write mode: overwrite, append, ignore, exclusive, backup, or update.
    header : bool
        Whether to write column names as the first line. Default is True.
    sep : str
        Field delimiter for the output file. Default is ','.
    quote : str
        Character used to quote fields. Default is '"'.
    quoteAll : bool
        Whether all values should be enclosed in quotes. Default is False.
    escape : str
        Character used to escape sep and quote when needed. Default is '\\'.
    timestampFormat : str
        Date format for datetime objects. Default is '%Y-%m-%dT%H:%M:%S.%f'.
    lineSep : str
        Character used as line separator. Default is os.linesep.
    compression : Optional[Literal["infer", "gzip", "bz2", "zip", "xz", "zstd", "tar"]]
        Compression to use for the output data. Default is None.

    See Also
    --------
    For more details on the CSV parameters, refer to the PandasCsvBufferWriter class documentation.
    """

    buffer_writer: Optional[PandasCsvBufferWriter] = Field(default=None, validate_default=False)

    @model_validator(mode="after")
    def set_up_buffer_writer(self) -> "SendCsvToSftp":
        """Set up the buffer writer, passing all CSV related options to it."""
        self.buffer_writer = PandasCsvBufferWriter(**self.get_options(options_type="koheesio_pandas_buffer_writer"))
        return self

    def execute(self) -> SFTPWriter.Output:
        SFTPWriter.execute(self)


class SendJsonToSftp(PandasJsonBufferWriter, SFTPWriter):
    """
    Write a DataFrame to an SFTP server as a JSON file.

    This class uses the PandasJsonBufferWriter to generate the JSON data and the SFTPWriter to write the data to the
    SFTP server.

    Example
    -------
    ```python
    from koheesio.spark.writers import SendJsonToSftp

    writer = SendJsonToSftp(
        # SFTP Parameters (Inherited from SFTPWriter)
        host="sftp.example.com",
        port=22,
        username="user",
        password="password",
        path="/path/to/folder",
        file_name="file.json.gz",
        # JSON Parameters (Inherited from PandasJsonBufferWriter)
        orient="records",
        date_format="iso",
        double_precision=2,
        date_unit="ms",
        lines=False,
        compression="gzip",
        index=False,
    )

    writer.write(df)
    ```

    In this example, the DataFrame `df` is written to the file `file.json.gz` in the folder `/path/to/folder` on the
    SFTP server. The file is written as a JSON file with gzip compression.

    Parameters
    ----------
    path : Union[str, Path]
        Path to the folder on the SFTP server.
    file_name : Optional[str]
        Name of the file, including extension. If not provided, expected to be part of the path.
    host : str
        SFTP Host.
    port : int
        SFTP Port.
    username : SecretStr
        SFTP Server Username.
    password : SecretStr
        SFTP Server Password.
    mode : SFTPWriteMode
        Write mode: overwrite, append, ignore, exclusive, backup, or update.
    orient : Literal["split", "records", "index", "columns", "values", "table"]
        Format of the JSON string. Default is 'records'.
    lines : bool
        If True, output is one JSON object per line. Only used when orient='records'. Default is True.
    date_format : Literal["iso", "epoch"]
        Type of date conversion. Default is 'iso'.
    double_precision : int
        Decimal places for encoding floating point values. Default is 10.
    force_ascii : bool
        If True, encoded string is ASCII. Default is True.
    compression : Optional[Literal["gzip"]]
        Compression to use for output data. Default is None.

    See Also
    --------
    For more details on the JSON parameters, refer to the PandasJsonBufferWriter class documentation.
    """

    buffer_writer: Optional[PandasJsonBufferWriter] = Field(default=None, validate_default=False)

    @model_validator(mode="after")
    def set_up_buffer_writer(self) -> "SendJsonToSftp":
        """Set up the buffer writer, passing all JSON related options to it."""
        self.buffer_writer = PandasJsonBufferWriter(
            **self.get_options(), compression=self.compression, columns=self.columns
        )
        return self

    def execute(self) -> SFTPWriter.Output:
        SFTPWriter.execute(self)



"""
This module defines writers to write Slowly Changing Dimension (SCD) Type 2 data to a Delta table.

Slowly Changing Dimension (SCD) is a technique used in data warehousing to handle changes to dimension data over time.
SCD Type 2 is one of the most common types of SCD, where historical changes are tracked by creating new records for each change.

Koheesio is a powerful data processing framework that provides advanced capabilities for working with Delta tables in Apache Spark.
It offers a convenient and efficient way to handle SCD Type 2 operations on Delta tables.

To learn more about Slowly Changing Dimension and SCD Type 2, you can refer to the following resources:
- [Slowly Changing Dimension (SCD) - Wikipedia](https://en.wikipedia.org/wiki/Slowly_changing_dimension)

By using Koheesio, you can benefit from its efficient merge logic, support for SCD Type 2 and SCD Type 1 attributes,
and seamless integration with Delta tables in Spark.

"""

from typing import List, Optional
from logging import Logger

from delta.tables import DeltaMergeBuilder, DeltaTable

from pydantic import InstanceOf

from pyspark.sql import functions as f
from pyspark.sql.types import DateType, TimestampType

from koheesio.models import Field
from koheesio.spark import Column, DataFrame, SparkSession
from koheesio.spark.delta import DeltaTableStep
from koheesio.spark.functions import current_timestamp_utc
from koheesio.spark.writers import Writer
from koheesio.spark.writers.delta.utils import get_delta_table_for_name


class SCD2DeltaTableWriter(Writer):
    """
    A class used to write Slowly Changing Dimension (SCD) Type 2 data to a Delta table.

    Attributes
    ----------
    table : InstanceOf[DeltaTableStep]
        The table to merge to.
    merge_key : str
        The key used for merging data.
    include_columns : List[str]
        Columns to be merged. Will be selected from DataFrame. Default is all columns.
    exclude_columns : List[str]
        Columns to be excluded from DataFrame.
    scd2_columns : List[str]
        List of attributes for SCD2 type (track changes).
    scd2_timestamp_col : Optional[Column]
        Timestamp column for SCD2 type (track changes). Default to current_timestamp.
    scd1_columns : List[str]
        List of attributes for SCD1 type (just update).
    meta_scd2_struct_col_name : str
        SCD2 struct name.
    meta_scd2_effective_time_col_name : str
        Effective col name.
    meta_scd2_is_current_col_name : str
        Current col name.
    meta_scd2_end_time_col_name : str
        End time col name.
    target_auto_generated_columns : List[str]
        Auto generated columns from target Delta table. Will be used to exclude from merge logic.
    """

    table: InstanceOf[DeltaTableStep] = Field(..., description="The table to merge to")
    merge_key: str = Field(..., description="Merge key")
    include_columns: List[str] = Field(
        default_factory=list, description="Columns to be merged. Will be selected from DataFrame.Default is all columns"
    )
    exclude_columns: List[str] = Field(default_factory=list, description="Columns to be excluded from DataFrame")
    scd2_columns: List[str] = Field(
        default_factory=list, description="List of attributes for scd2 type (track changes)"
    )
    scd2_timestamp_col: Column = Field(
        default=None,
        description="Timestamp column for SCD2 type (track changes). Default to current_timestamp",
    )
    scd1_columns: List[str] = Field(default_factory=list, description="List of attributes for scd1 type (just update)")
    meta_scd2_struct_col_name: str = Field(default="_scd2", description="SCD2 struct name")
    meta_scd2_effective_time_col_name: str = Field(default="effective_time", description="Effective col name")
    meta_scd2_is_current_col_name: str = Field(default="is_current", description="Current col name")
    meta_scd2_end_time_col_name: str = Field(default="end_time", description="End time col name")
    target_auto_generated_columns: List[str] = Field(
        default_factory=list,
        description="Auto generated columns from target Delta table. Will be used to exclude from merge logic",
    )

    @staticmethod
    def _prepare_attr_clause(attrs: List[str], src_alias: str, dest_alias: str) -> Optional[str]:
        """
        Prepare an attribute clause for SQL query.

        This method takes a list of attributes and two aliases, and prepares a clause for SQL query.
        The clause checks if the attributes in the source alias and destination alias are not equal.

        Parameters
        ----------
        attrs : List[str]
            List of attributes to be checked.
        src_alias : str
            Alias for the source table in the SQL query.
        dest_alias : str
            Alias for the destination table in the SQL query.

        Returns
        -------
        Optional[str]
            The prepared SQL clause if attributes are provided, None otherwise.

        """
        attr_clause = None

        if attrs:
            attr_clause = list(map(lambda attr: f"NOT ({src_alias}.{attr} <=> {dest_alias}.{attr})", attrs))
            attr_clause = " OR ".join(attr_clause)  # type: ignore

        return attr_clause

    @staticmethod
    def _scd2_timestamp(spark: SparkSession, scd2_timestamp_col: Optional[Column] = None, **_kwargs) -> Column:
        """
        Generate a SCD2 timestamp column.

        This method generates a SCD2 timestamp column. If `scd2_timestamp_col` is provided, it is cast to TimestampType.
        Otherwise, the current UTC timestamp is used.

        Parameters
        ----------
        spark : SparkSession
            The SparkSession to use.
        scd2_timestamp_col : Optional[Column], default is None
            The column to cast to TimestampType. If None, the current UTC timestamp is used.

        Returns
        -------
        Column
            The generated SCD2 timestamp column.

        """
        scd2_timestamp = (
            current_timestamp_utc(spark=spark)
            if scd2_timestamp_col is None
            else scd2_timestamp_col.cast(TimestampType())
        )
        return scd2_timestamp

    @staticmethod
    def _scd2_end_time(meta_scd2_end_time_col: str, **_kwargs: dict) -> Column:
        """
        Generate a SCD2 end time column.

        This method generates a SCD2 end time column based on the provided `meta_scd2_end_time_col`.
        If the `__meta_scd2_system_merge_action` is 'UC' and `__meta_scd2_rn` is 2, the `__meta_scd2_timestamp` is used.
        Otherwise, the value from the target's `meta_scd2_end_time_col` is used.

        Parameters
        ----------
        meta_scd2_end_time_col : str
            The column name in the target table that holds the SCD2 end time.

        Returns
        -------
        Column
            The generated SCD2 end time column.

        """
        scd2_end_time = f.expr(
            "CASE WHEN __meta_scd2_system_merge_action='UC' AND cross.__meta_scd2_rn=2 THEN __meta_scd2_timestamp "
            f"     ELSE tgt.{meta_scd2_end_time_col} END"
        )

        return scd2_end_time

    @staticmethod
    def _scd2_effective_time(meta_scd2_effective_time_col: str, **_kwargs: dict) -> Column:
        """
        Generate a SCD2 effective time column.

        This method generates a SCD2 effective time column based on the provided `meta_scd2_effective_time_col`.
        If the `__meta_scd2_system_merge_action` is 'UC' and `__meta_scd2_rn` is 1, the `__meta_scd2_timestamp` is used.
        Otherwise, the value from the target's `meta_scd2_effective_time_col` is used,
        if it exists, or `__meta_scd2_timestamp` is used.

        Parameters
        ----------
        meta_scd2_effective_time_col : str
            The column name in the target table that holds the SCD2 effective time.

        Returns
        -------
        Column
            The generated SCD2 effective time column.

        """
        scd2_effective_time = f.when(
            f.expr("__meta_scd2_system_merge_action='UC' and cross.__meta_scd2_rn=1"),
            f.col("__meta_scd2_timestamp"),
        ).otherwise(f.coalesce(meta_scd2_effective_time_col, "__meta_scd2_timestamp"))

        return scd2_effective_time

    @staticmethod
    def _scd2_is_current(**_kwargs: dict) -> Column:
        """
        Generate a SCD2 is_current column.

        This method generates a SCD2 is_current column. If the `__meta_scd2_system_merge_action` is 'UC'
        and `__meta_scd2_rn` is 2, the method returns False. Otherwise, it returns True.

        Returns
        -------
        Column
            The generated SCD2 is_current column.

        """
        scd2_is_current = f.expr(
            "CASE WHEN __meta_scd2_system_merge_action='UC' AND cross.__meta_scd2_rn=2 THEN False ELSE True END"
        )

        return scd2_is_current

    def _prepare_staging(
        self,
        df: DataFrame,
        delta_table: DeltaTable,
        merge_action_logic: Column,
        meta_scd2_is_current_col: str,
        columns_to_process: List[str],
        src_alias: str,
        dest_alias: str,
        cross_alias: str,
        **_kwargs: dict,
    ) -> DataFrame:
        """
        Prepare a DataFrame for staging.

        This method prepares a DataFrame for staging by selecting the necessary columns, joining with the delta table,
        adding a merge action column, filtering based on the merge action, and cross joining with a DataFrame of two
        rows.

        Parameters
        ----------
        df : DataFrame
            The DataFrame to prepare for staging.
        delta_table : DeltaTable
            The DeltaTable to join with.
        merge_key : str
            The key to use for the join.
        merge_action_logic : Column
            The logic to use for the merge action.
        meta_scd2_is_current_col : str
            The column that indicates whether a row is current.
        columns_to_process : List[str]
            The columns to select from the DataFrame.
        src_alias : str
            The alias to use for the DataFrame.
        dest_alias : str
            The alias to use for the DeltaTable.
        cross_alias : str
            The alias to use for the cross join.

        Returns
        -------
        DataFrame
            The prepared DataFrame.

        """
        df = (
            df.select(*columns_to_process, "__meta_scd2_timestamp")
            .alias(src_alias)
            .join(
                other=delta_table.toDF()
                .filter(f.col(meta_scd2_is_current_col).eqNullSafe(f.lit(True)))
                .alias(dest_alias),
                on=self.merge_key,
                how="left",
            )
            .withColumn("__meta_scd2_system_merge_action", merge_action_logic)
            .filter("__meta_scd2_system_merge_action IS NOT NULL")
            .crossJoin(other=self.spark.sql("SELECT 1 as __meta_scd2_rn UNION ALL SELECT 2").alias(cross_alias))
            # Filter cross joined data so that we have one row for U
            # and another for I in case of closing SCD2
            # and keep just one for SCD1 or NEW row
            .filter(f.expr("__meta_scd2_system_merge_action='UC' OR cross.__meta_scd2_rn=1"))
        )

        return df

    @staticmethod
    def _preserve_existing_target_values(
        df: DataFrame,
        meta_scd2_struct_col_name: str,
        target_auto_generated_columns: List[str],
        src_alias: str,
        cross_alias: str,
        dest_alias: str,
        logger: Logger,
        **_kwargs: dict,
    ) -> DataFrame:
        """
        Preserve existing target values in the DataFrame.

        This method preserves existing target values in the DataFrame. It skips certain columns and for
        the remaining columns,if the `__meta_scd2_system_merge_action` is 'UC' and `__meta_scd2_rn`
        is 2, it keeps the value from the destination table.
        Otherwise, it keeps the value from the source table.

        Parameters
        ----------
        df : DataFrame
            The DataFrame to process.
        meta_scd2_struct_col_name : str
            The column name of the SCD2 struct.
        target_auto_generated_columns : List[str]
            The list of auto-generated columns in the target.
        src_alias : str
            The alias to use for the source DataFrame.
        cross_alias : str
            The alias to use for the cross join.
        dest_alias : str
            The alias to use for the destination DataFrame.
        logger : Logger
            The logger to use.

        Returns
        -------
        DataFrame
            The processed DataFrame.

        """
        skipped_columns = [
            meta_scd2_struct_col_name,
            "__meta_scd2_system_merge_action",
            "__meta_scd2_timestamp",
            "__meta_scd2_rn",
        ] + target_auto_generated_columns

        for c in set(df.columns):
            if c not in skipped_columns:
                df = (
                    df.withColumn(
                        f"newly_{c}",
                        f.when(
                            f.col("__meta_scd2_system_merge_action").eqNullSafe(f.lit("UC"))
                            & f.col(f"{cross_alias}.__meta_scd2_rn").eqNullSafe(f.lit(2)),
                            f.col(f"{dest_alias}.{c}"),
                        ).otherwise(f.col(f"{src_alias}.{c}")),
                    )
                    .drop(f.col(f"{src_alias}.{c}"))
                    .drop(f.col(f"{dest_alias}.{c}"))
                    .withColumnRenamed(f"newly_{c}", c)
                )

        logger.debug(f"Keep old values for columns:{df.columns}")

        return df

    @staticmethod
    def _add_scd2_columns(
        df: DataFrame,
        meta_scd2_struct_col_name: str,
        meta_scd2_effective_time_col_name: str,
        meta_scd2_end_time_col_name: str,
        meta_scd2_is_current_col_name: str,
        **_kwargs: dict,
    ) -> DataFrame:
        """
        Add SCD2 columns to the DataFrame.

        This method adds SCD2 columns to the DataFrame.
        Parameters
        ----------
        df : DataFrame
            The DataFrame to add the SCD2 columns to.
        meta_scd2_struct_col_name : str
            The name of the struct column to add.
        meta_scd2_effective_time_col_name : str
            The name of the effective time column inside the struct.
        meta_scd2_end_time_col_name : str
            The name of the end time column inside the struct.
        meta_scd2_is_current_col_name : str
            The name of the is_current column inside the struct.

        Returns
        -------
        DataFrame
            The DataFrame with the added SCD2 columns.

        """
        df = df.withColumn(
            meta_scd2_struct_col_name,
            f.struct(
                f.col("__meta_scd2_effective_time").alias(meta_scd2_effective_time_col_name),
                f.col("__meta_scd2_end_time").alias(meta_scd2_end_time_col_name),
                f.col("__meta_scd2_is_current").alias(meta_scd2_is_current_col_name),
            ),
        ).drop(
            "__meta_scd2_end_time",
            "__meta_scd2_is_current",
            "__meta_scd2_effective_time",
            "__meta_scd2_timestamp",
            "__meta_scd2_system_merge_action",
            "__meta_scd2_rn",
        )

        return df

    def _prepare_merge_builder(
        self,
        delta_table: DeltaTable,
        dest_alias: str,
        staged: DataFrame,
        merge_key: str,
        columns_to_process: List[str],
        meta_scd2_effective_time_col: str,
        **_kwargs: dict,
    ) -> DeltaMergeBuilder:
        """
        Prepare a DeltaMergeBuilder for merging data.

        This method prepares a DeltaMergeBuilder for merging data. It sets up the merge conditions and specifies
        the columns to update or insert when a match is found or not found, respectively.

        Parameters
        ----------
        delta_table : DeltaTable
            The DeltaTable to merge with.
        dest_alias : str
            The alias to use for the DeltaTable.
        staged : DataFrame
            The DataFrame to merge.
        merge_key : str
            The key to use for the merge.
        columns_to_process : List[str]
            The columns to update or insert.
        meta_scd2_effective_time_col : str
            The effective time column to use in the merge condition.

        Returns
        -------
        DeltaMergeBuilder
            The prepared DeltaMergeBuilder.

        """
        when_matched_columns = {c: f"src.{c}" for c in columns_to_process + [self.meta_scd2_struct_col_name]}

        merge_builder = (
            delta_table.alias(dest_alias)
            .merge(
                source=staged.alias("src"),
                condition=f"src.{merge_key}={dest_alias}.{merge_key}"
                f" AND src.{meta_scd2_effective_time_col}={dest_alias}.{meta_scd2_effective_time_col}",
            )
            .whenMatchedUpdate(set=when_matched_columns)  # type: ignore
            .whenNotMatchedInsert(values=when_matched_columns)  # type: ignore
        )

        return merge_builder

    def execute(self) -> None:
        """
        Execute the SCD Type 2 operation.

        This method executes the SCD Type 2 operation on the DataFrame.
        It validates the existing Delta table, prepares the merge conditions, stages the data,
        and then performs the merge operation.

        Raises
        ------
        TypeError
            If the scd2_timestamp_col is not of date or timestamp type.
            If the source DataFrame is missing any of the required merge columns.

        """
        self.df: DataFrame
        self.spark: SparkSession
        delta_table = get_delta_table_for_name(spark_session=self.spark, table_name=self.table.table_name)
        src_alias, cross_alias, dest_alias = "src", "cross", "tgt"

        # Prepare required merge columns
        required_merge_columns = [self.merge_key]

        if self.scd2_columns:
            required_merge_columns += self.scd2_columns

        if self.scd1_columns:
            required_merge_columns += self.scd1_columns

        if not all(c in self.df.columns for c in required_merge_columns):
            missing_columns = [c for c in required_merge_columns if c not in self.df.columns]
            raise TypeError(f"The source DataFrame is missing the columns: {missing_columns!r}")

        # Check that required columns are present in the source DataFrame
        if self.scd2_timestamp_col is not None:
            timestamp_col_type = self.df.select(self.scd2_timestamp_col).schema.fields[0].dataType

            if not isinstance(timestamp_col_type, (DateType, TimestampType)):
                raise TypeError(
                    f"The scd2_timestamp_col '{self.scd2_timestamp_col}' must be of date "
                    f"or timestamp type.Current type is {timestamp_col_type}"
                )

        # Prepare columns to process
        include_columns = self.include_columns if self.include_columns else self.df.columns
        exclude_columns = self.exclude_columns
        columns_to_process = [c for c in include_columns if c not in exclude_columns]

        # Constructing column names for SCD2 attributes
        meta_scd2_is_current_col = f"{self.meta_scd2_struct_col_name}.{self.meta_scd2_is_current_col_name}"
        meta_scd2_effective_time_col = f"{self.meta_scd2_struct_col_name}.{self.meta_scd2_effective_time_col_name}"
        meta_scd2_end_time_col = f"{self.meta_scd2_struct_col_name}.{self.meta_scd2_end_time_col_name}"

        # Constructing system merge action logic
        system_merge_action = f"CASE WHEN tgt.{self.merge_key} is NULL THEN 'I' "

        if updates_attrs_scd2 := self._prepare_attr_clause(
            attrs=self.scd2_columns, src_alias=src_alias, dest_alias=dest_alias
        ):
            system_merge_action += f" WHEN {updates_attrs_scd2} THEN 'UC' "

        if updates_attrs_scd1 := self._prepare_attr_clause(
            attrs=self.scd1_columns, src_alias=src_alias, dest_alias=dest_alias
        ):
            system_merge_action += f" WHEN {updates_attrs_scd1} THEN 'U' "

        system_merge_action += " ELSE NULL END"

        # Prepare the staged DataFrame
        staged = (
            self.df.withColumn(
                "__meta_scd2_timestamp",
                self._scd2_timestamp(scd2_timestamp_col=self.scd2_timestamp_col, spark=self.spark),
            )
            .transform(
                func=self._prepare_staging,
                delta_table=delta_table,
                merge_action_logic=f.expr(system_merge_action),
                meta_scd2_is_current_col=meta_scd2_is_current_col,
                columns_to_process=columns_to_process,
                src_alias=src_alias,
                dest_alias=dest_alias,
                cross_alias=cross_alias,
            )
            .transform(
                func=self._preserve_existing_target_values,
                meta_scd2_struct_col_name=self.meta_scd2_struct_col_name,
                target_auto_generated_columns=self.target_auto_generated_columns,
                src_alias=src_alias,
                cross_alias=cross_alias,
                dest_alias=dest_alias,
                logger=self.log,
            )
            .withColumn("__meta_scd2_end_time", self._scd2_end_time(meta_scd2_end_time_col=meta_scd2_end_time_col))
            .withColumn("__meta_scd2_is_current", self._scd2_is_current())
            .withColumn(
                "__meta_scd2_effective_time",
                self._scd2_effective_time(meta_scd2_effective_time_col=meta_scd2_effective_time_col),
            )
            .transform(
                func=self._add_scd2_columns,
                meta_scd2_struct_col_name=self.meta_scd2_struct_col_name,
                meta_scd2_effective_time_col_name=self.meta_scd2_effective_time_col_name,
                meta_scd2_end_time_col_name=self.meta_scd2_end_time_col_name,
                meta_scd2_is_current_col_name=self.meta_scd2_is_current_col_name,
            )
        )

        self._prepare_merge_builder(
            delta_table=delta_table,
            dest_alias=dest_alias,
            staged=staged,
            merge_key=self.merge_key,
            columns_to_process=columns_to_process,
            meta_scd2_effective_time_col=meta_scd2_effective_time_col,
        ).execute()



"""
This module contains the implementation of the `DownloadFileStep` class, which is responsible for downloading files
from a given URL and saving them to a specified local directory path.

It supports various file write modes such as overwrite, append, ignore, exclusive, and backup.

Classes
-------
FileWriteMode
    Enum representing different file write modes.
DownloadFileStep
    Class for downloading files with support for different write modes.
"""

from __future__ import annotations

from typing import Union
from enum import Enum
from pathlib import Path
import time

from koheesio import StepOutput
from koheesio.models import DirectoryPath, Field, FilePath, field_validator
from koheesio.steps.http import HttpGetStep


# pylint: disable=E1101
class FileWriteMode(str, Enum):
    """
    The different write modes for the DownloadFileStep.

    ### OVERWRITE:
    * If the file exists, it will be overwritten.
    * If it does not exist, a new file will be created.

    ### APPEND:
    * If the file exists, the new data will be appended to it.
    * If it does not exist, a new file will be created.

    ### IGNORE:
    * If the file exists, the method will return without writing anything.
    * If it does not exist, a new file will be created.

    ### EXCLUSIVE:
    * If the file exists, an error will be raised.
    * If it does not exist, a new file will be created.

    ### BACKUP:
    * If the file exists, a backup will be created and the original file will be overwritten.
    * If it does not exist, a new file will be created.
    """

    OVERWRITE = "overwrite"
    APPEND = "append"
    IGNORE = "ignore"
    EXCLUSIVE = "exclusive"
    BACKUP = "backup"

    @classmethod
    def from_string(cls, mode: str) -> FileWriteMode:
        """Return the FileWriteMode for the given string.

        Parameters
        ----------
        mode : str
            The string representation of the FileWriteMode.

        Returns
        -------
        FileWriteMode
            The FileWriteMode enum corresponding to the given string
        """
        return cls[mode.upper()]

    @property
    def write_mode(self) -> str:
        """Return the write mode for the given FileWriteMode."""
        if self in {FileWriteMode.OVERWRITE, FileWriteMode.BACKUP, FileWriteMode.EXCLUSIVE}:
            # OVERWRITE, BACKUP, and EXCLUSIVE modes set the file to be written from the beginning
            return "wb"
        if self == FileWriteMode.APPEND:
            # APPEND mode sets the file to be written from the end
            return "ab"


class DownloadFileStep(HttpGetStep):
    """
    Downloads a file from the given URL and saves it to the specified download path.

    Example
    -------
    ```python
    URL = "http://example.com/testfile.txt"
    download_path = Path("downloads")
    step = DownloadFileStep(
        url=URL, download_path=download_path, mode="ignore"
    )
    step.execute()
    ```

    In the above example, the file `testfile.txt` will be downloaded from the URL `http://example.com/testfile.txt` to
    the `downloads` directory.

    Parameters
    ----------
    url : str
        The URL to download the file from.
    download_path : str
        The local directory path where the file will be downloaded.
    chunk_size : int, optional, default=8192
        The size (in bytes) of the chunks to download the file in, must be greater than 16.
    mode : FileWriteMode, optional, default=FileWriteMode.OVERWRITE
        Write mode: overwrite, append, ignore, exclusive, or backup.
        See the docstring of `FileWriteMode` for more details.
    """

    download_path: DirectoryPath = Field(
        ..., description="The local directory path where the file will be downloaded to."
    )
    chunk_size: int = Field(
        8192,
        ge=16,
        description="The size (in bytes) of the chunks to download the file in, must be greater than or equal to 16.",
    )
    mode: FileWriteMode = Field(
        default=FileWriteMode.OVERWRITE,
        description="Write mode: overwrite, append, ignore, exclusive, backup, or update.",
    )

    class Output(StepOutput):
        download_file_path: FilePath = Field(..., description="The full path where the file was downloaded to.")

    @field_validator("mode")
    def validate_mode(cls, v: Union[str, FileWriteMode]) -> FileWriteMode:
        """Ensure that the mode is a valid FileWriteMode."""
        return FileWriteMode.from_string(v) if isinstance(v, str) else v

    def should_write_file(self, _filepath: Path, _filename: str) -> bool:
        """
        Determine if the file should be written based on the write mode.

        Parameters
        ----------
        _filepath : Path
            The path of the file to be written.
        _filename : str
            The name of the file to be written.

        Returns
        -------
        bool
            True if the file should be written, False otherwise.
        """
        _mode = self.mode

        # OVERWRITE and APPEND modes will write the file irrespective of whether it exists or not
        if _filepath.exists() and _mode not in {FileWriteMode.OVERWRITE, FileWriteMode.APPEND}:
            if _mode == FileWriteMode.IGNORE:
                # If the file exists in IGNORE mode, return False
                self.log.info(f"File {_filepath} already exists. Ignoring {_filename} based on IGNORE mode.")
                self.output.download_file_path = _filepath
                return False

            elif _mode == FileWriteMode.EXCLUSIVE:
                raise FileExistsError(
                    f"File {_filepath} already exists. Cannot write to {_filename} based on EXCLUSIVE mode."
                )

            elif _mode == FileWriteMode.BACKUP:
                # In BACKUP mode, we first create a timestamped backup before overwriting the existing file.
                file_to_be_backed_up = _filepath
                backup_path = _filepath.with_suffix(f"{_filepath.suffix}.{int(time.time())}.bak")
                # create the backup
                self.log.info(f"Creating backup of {_filename} as {backup_path}...")
                file_to_be_backed_up.rename(backup_path)

        return True

    def execute(self) -> Output:
        """
        Executes the file download process, handling different write modes, and saving the file to the specified path.

        Returns
        -------
        Output
            An instance of the Output class containing the path where the file was downloaded.
        """
        _filename = Path(self.url).name
        _filepath = self.download_path / _filename
        _write_mode = self.mode.write_mode

        # Check if the file should be written based on the given mode
        if not self.should_write_file(_filepath, _filename):
            return self.output

        # Create the download path if it does not exist
        self.output.download_file_path = _filepath
        self.output.download_file_path.touch(exist_ok=True)

        with self._request(stream=True) as response:  # type: ignore
            with self.output.download_file_path.open(mode=_write_mode) as f:  # type: ignore
                for chunk in response.iter_content(chunk_size=self.chunk_size):
                    self.log.debug(f"Downloading chunk of size {len(chunk)}")
                    self.log.debug(f"Downloaded {f.tell()} bytes")
                    self.log.debug(f"Writing to file {self.output.download_file_path}")
                    f.write(chunk)



import pytest

from koheesio.spark.readers.hana import HanaReader

pytestmark = pytest.mark.spark


class TestHanaReader:
    common_options = {
        "url": "url",
        "user": "user",
        "password": "password",
        "dbtable": "table",
    }

    def test_get_options(self):
        hana = HanaReader(**self.common_options)
        o = hana.get_options()

        assert o["driver"] == "com.sap.db.jdbc.Driver"
        assert o["fetchsize"] == 2000
        assert o["numPartitions"] == 10



"""
Snowflake steps and tasks for Koheesio
"""

from koheesio.integrations.spark.snowflake import (
    AddColumn,
    CreateOrReplaceTableFromDataFrame,
    DbTableQuery,
    GetTableSchema,
    GrantPrivilegesOnFullyQualifiedObject,
    GrantPrivilegesOnObject,
    GrantPrivilegesOnTable,
    GrantPrivilegesOnView,
    Query,
    RunQuery,
    SnowflakeBaseModel,
    SnowflakeReader,
    SnowflakeStep,
    SnowflakeTableStep,
    SnowflakeTransformation,
    SnowflakeWriter,
    SynchronizeDeltaToSnowflakeTask,
    SyncTableAndDataFrameSchema,
    TableExists,
)

__all__ = [
    "AddColumn",
    "CreateOrReplaceTableFromDataFrame",
    "DbTableQuery",
    "GetTableSchema",
    "GrantPrivilegesOnFullyQualifiedObject",
    "GrantPrivilegesOnObject",
    "GrantPrivilegesOnTable",
    "GrantPrivilegesOnView",
    "Query",
    "RunQuery",
    "SnowflakeBaseModel",
    "SnowflakeReader",
    "SnowflakeStep",
    "SnowflakeTableStep",
    "SnowflakeTransformation",
    "SnowflakeWriter",
    "SyncTableAndDataFrameSchema",
    "SynchronizeDeltaToSnowflakeTask",
    "TableExists",
]



"""Module holding re-usable test utilities for Snowflake modules"""

from typing import Generator
from unittest.mock import MagicMock, patch

# safe import pytest fixture
try:
    import pytest
except (ImportError, ModuleNotFoundError):
    pytest = MagicMock()


@pytest.fixture(scope="function")
def mock_query() -> Generator:
    """Mock the query execution for SnowflakeRunQueryPython

    This can be used to test the query execution without actually connecting to Snowflake.

    Example
    -------
    ```python
    def test_execute(self, mock_query):
        # Arrange
        query = "SELECT * FROM two_row_table"
        mock_query.expected_data = [("row1",), ("row2",)]

        # Act
        instance = SnowflakeRunQueryPython(
            **COMMON_OPTIONS, query=query, account="42"
        )
        instance.execute()

        # Assert
        mock_query.assert_called_with(query)
        assert instance.output.results == mock_query.expected_data
    ```

    In this example, we are using the mock_query fixture to test the execution of a query.
    - We set the expected data to a known value by setting `mock_query.expected_data`,
    - Then, we execute the query.
    - We then assert that the query was called with the expected query by using `mock_query.assert_called_with` and
        that the results are as expected.
    """
    with patch("koheesio.integrations.snowflake.SnowflakeRunQueryPython.conn", new_callable=MagicMock) as mock_conn:
        mock_cursor = MagicMock()
        mock_conn.__enter__.return_value.execute_string.return_value = [mock_cursor]

        class MockQuery:
            def __init__(self) -> None:
                self.mock_conn = mock_conn
                self.mock_cursor = mock_cursor
                self._expected_data: list = []

            def assert_called_with(self, query: str) -> None:
                self.mock_conn.__enter__.return_value.execute_string.assert_called_once_with(query)
                self.mock_cursor.fetchall.return_value = self.expected_data

            @property
            def expected_data(self) -> list:
                return self._expected_data

            @expected_data.setter
            def expected_data(self, data: list) -> None:
                self._expected_data = data
                self.set_expected_data()

            def set_expected_data(self) -> None:
                self.mock_cursor.fetchall.return_value = self.expected_data

        mock_query_instance = MockQuery()
        yield mock_query_instance



"""unit tests for Trigger"""

import pytest

from koheesio.spark.writers.stream import Trigger

pytestmark = pytest.mark.spark


@pytest.mark.parametrize(
    "args, expected",
    [
        # key can be written either in lowerCamelCase or snake_case
        ({"processingTime": "5 seconds"}, {"processingTime": "5 seconds"}),
        ({"processingTime": "5 seconds"}, {"processingTime": "5 seconds"}),
        ({"processing_time": "5 hours"}, {"processingTime": "5 hours"}),
        ({"processingTime": "4 minutes"}, {"processingTime": "4 minutes"}),
        # boolean values can be passed as a string or as a boolean (case-insensitive)
        ({"once": True}, {"once": True}),
        ({"once": "true"}, {"once": True}),
        ({"once": "TrUe"}, {"once": True}),
        ({"once": "TRUE"}, {"once": True}),
        # all keys should work
        ({"available_now": "true"}, {"availableNow": True}),
        ({"availableNow": True}, {"availableNow": True}),
        ({"continuous": "3 hours"}, {"continuous": "3 hours"}),
        # from a string
        ("continuous='5 seconds'", {"continuous": "5 seconds"}),
        ("processingTime=4 minutes", {"processingTime": "4 minutes"}),
    ],
)
def test_from_any(args, expected):
    """Test that Trigger.from_any can handle different input types
    Inherently also tests that Trigger.from_string works as expected
    """
    trigger = Trigger.from_any(args)
    actual = trigger.value
    assert actual == expected


@pytest.mark.parametrize(
    "args, expected",
    [
        # valid values, but should fail the validation check of the class
        ({"availableNow": False}, ValueError),
        ({"continuous": True}, ValueError),
        ({"once": False}, ValueError),
        ({"once": "string should fail"}, ValueError),
        # invalid values
        ({"something_random": "false"}, ValueError),
        # multiple keys should trigger a ValueError
        ({"availableNow": True, "continuous": "5 seconds"}, ValueError),
    ],
)
def test_from_any_unhappy(args, expected):
    """Test that Trigger.from_any raises a ValueError when an invalid type is passed"""
    with pytest.raises(expected):
        Trigger.from_any(args)


def test_execute():
    """Test that Trigger.execute returns a valid Spark Trigger"""
    trigger = Trigger(processing_time="5 seconds")
    actual_value = trigger.value
    actual_execute = trigger.execute()
    assert actual_value == {"processingTime": "5 seconds"} == actual_execute.value



from koheesio.spark.utils.common import (
    SPARK_MINOR_VERSION,
    SparkDatatype,
    check_if_pyspark_connect_is_supported,
    check_if_pyspark_connect_module_is_available,
    get_column_name,
    get_spark_minor_version,
    import_pandas_based_on_pyspark_version,
    on_databricks,
    schema_struct_to_schema_str,
    show_string,
    spark_data_type_is_array,
    spark_data_type_is_numeric,
)

__all__ = [
    "SparkDatatype",
    "import_pandas_based_on_pyspark_version",
    "on_databricks",
    "schema_struct_to_schema_str",
    "spark_data_type_is_array",
    "spark_data_type_is_numeric",
    "show_string",
    "get_spark_minor_version",
    "SPARK_MINOR_VERSION",
    "check_if_pyspark_connect_module_is_available",
    "check_if_pyspark_connect_is_supported",
    "get_column_name",
]



"""The Writer class is used to write the DataFrame to a target."""

from typing import Optional
from abc import ABC, abstractmethod
from enum import Enum

from koheesio.models import Field
from koheesio.spark import DataFrame, SparkStep


# TODO: Investigate if we can clean various OutputModes into a more streamlined structure
class BatchOutputMode(str, Enum):
    """For Batch:

    - append: Append the contents of the DataFrame to the output table, default option in Koheesio.
    - overwrite: overwrite the existing data.
    - ignore: ignore the operation (i.e. no-op).
    - error or errorifexists: throw an exception at runtime.
    - merge: update matching data in the table and insert rows that do not exist.
    - merge_all: update matching data in the table and insert rows that do not exist.
    """

    # Batch only modes
    APPEND = "append"
    OVERWRITE = "overwrite"
    IGNORE = "ignore"
    ERROR = "error"
    ERRORIFEXISTS = "error"
    MERGE_ALL = "merge_all"
    MERGEALL = "merge_all"
    MERGE = "merge"


class StreamingOutputMode(str, Enum):
    """For Streaming:

    - append: only the new rows in the streaming DataFrame will be written to the sink.
    - complete: all the rows in the streaming DataFrame/Dataset will be written to the sink every time there are some
       updates.
    - update: only the rows that were updated in the streaming DataFrame/Dataset will be written to the sink every time
       there are some updates. If the query doesn't contain aggregations, it will be equivalent to append mode.
    """

    # Streaming only modes
    APPEND = "append"
    COMPLETE = "complete"
    UPDATE = "update"


class Writer(SparkStep, ABC):
    """The Writer class is used to write the DataFrame to a target."""

    df: Optional[DataFrame] = Field(default=None, description="The Spark DataFrame", exclude=True)
    format: str = Field(default="delta", description="The format of the output")

    @property
    def streaming(self) -> bool:
        """Check if the DataFrame is a streaming DataFrame or not."""
        if not self.df:
            raise RuntimeError("No valid Dataframe was passed")

        return self.df.isStreaming

    @abstractmethod
    def execute(self) -> SparkStep.Output:
        """Execute on a Writer should handle writing of the self.df (input) as a minimum"""
        # self.df  # input dataframe
        ...

    def write(self, df: Optional[DataFrame] = None) -> SparkStep.Output:
        """Write the DataFrame to the output using execute() and return the output.

        If no DataFrame is passed, the self.df will be used.
        If no self.df is set, a RuntimeError will be thrown.
        """
        self.df = df or self.df
        if not self.df:
            raise RuntimeError("No valid Dataframe was passed")
        self.execute()
        return self.output



from datetime import timedelta

import pytest

from koheesio.utils.date_time import DTInterval


@pytest.mark.parametrize(
    "input,expected",
    [
        (
            # check with all values
            "5 years 3 months 2 weeks 4 days 10 hours 42 minutes 4 seconds 20 milliseconds 200 microseconds",
            timedelta(
                days=int(4 + 14 + 3 * 30.44 + 5 * 365.25),
                seconds=4,
                microseconds=200,
                milliseconds=20,
                minutes=42,
                hours=10,
            ),
        ),
        (
            # check with aliases
            "7 years 4 months 1 weeks 3 days 8 hours 15 mins 30 secs 50 millis 100 micros",
            timedelta(
                days=int(3 + 7 + 4 * 30.44 + 7 * 365.25),
                seconds=30,
                microseconds=100,
                milliseconds=50,
                minutes=15,
                hours=8,
            ),
        ),
        (
            # check with singular values
            "1 year 1 month 1 week 1 day 1 hour 1 minute 1 second 1 millisecond 1 microsecond",
            timedelta(days=395 + 8, seconds=1, microseconds=1, milliseconds=1, minutes=1, hours=1),
        ),
        (
            # check with singular alias values
            "1 min 1 sec 1 milli 1 micro",
            timedelta(seconds=1, microseconds=1, milliseconds=1, minutes=1),
        ),
        (
            # test with week value
            "1 week",
            timedelta(weeks=1),
        ),
        ("5 years", timedelta(days=int(365.25 * 5))),
    ],
)
def test_dt_interval(input, expected):
    dt_interval = DTInterval(interval=input)
    assert dt_interval.to_timedelta == expected



"""Module for the DummyWriter class."""

from typing import Any, Dict, Union

from koheesio.models import Field, PositiveInt, field_validator
from koheesio.spark import DataFrame
from koheesio.spark.utils import show_string
from koheesio.spark.writers import Writer


class DummyWriter(Writer):
    """
    A simple DummyWriter that performs the equivalent of a df.show() on the given DataFrame and returns the first row of
    data as a dict.

    This Writer does not actually write anything to a source/destination, but is useful for debugging or testing
    purposes.

    Parameters
    ----------
    n : PositiveInt, optional, default=20
        Number of rows to show.
    truncate : bool | PositiveInt, optional, default=True
        If set to `True`, truncate strings longer than 20 chars by default.
        If set to a number greater than one, truncates long strings to length `truncate` and align cells right.
    vertical : bool, optional, default=False
        If set to `True`, print output rows vertically (one line per column value).
    """

    n: PositiveInt = Field(
        default=20,
        description="Number of rows to show.",
        gt=0,
    )
    truncate: Union[bool, PositiveInt] = Field(
        default=True,
        description="If set to ``True``, truncate strings longer than 20 chars by default."
        "If set to a number greater than one, truncates long strings to length ``truncate`` and align cells right.",
    )
    vertical: bool = Field(
        default=False,
        description="If set to ``True``, print output rows vertically (one line per column value).",
    )

    @field_validator("truncate")
    def int_truncate(cls, truncate_value: Union[int, bool]) -> int:
        """
        Truncate is either a bool or an int.

        Parameters:
        -----------
        truncate_value : int | bool, optional, default=True
            If int, specifies the maximum length of the string.
            If bool and True, defaults to a maximum length of 20 characters.

        Returns:
        --------
        int
            The maximum length of the string.

        """
        # Same logic as what is inside DataFrame.show()
        if isinstance(truncate_value, bool) and truncate_value is True:
            return 20  # default is 20 chars
        return int(truncate_value)  # otherwise 0, or whatever the user specified

    class Output(Writer.Output):
        """DummyWriter output"""

        head: Dict[str, Any] = Field(default=..., description="The first row of the DataFrame as a dict")
        df_content: str = Field(default=..., description="The content of the DataFrame as a string")

    def execute(self) -> Output:
        """Execute the DummyWriter"""
        # logs the equivalent of doing df.show()
        df_content = show_string(df=self.df, n=self.n, truncate=self.truncate, vertical=self.vertical)
        self.log.info(f"content of df that was passed to DummyWriter:\n{df_content}")

        self.output.head = self.df.head().asDict()
        self.output.df_content = df_content



from typing import List, Optional
import datetime

from delta import DeltaTable
from delta.tables import DeltaMergeBuilder
import pytest

from pydantic import Field

from pyspark.sql import Column
from pyspark.sql import functions as F
from pyspark.sql.types import Row

from koheesio.spark import DataFrame
from koheesio.spark.delta import DeltaTableStep
from koheesio.spark.functions import current_timestamp_utc
from koheesio.spark.utils import SPARK_MINOR_VERSION
from koheesio.spark.writers.delta.scd import SCD2DeltaTableWriter
from koheesio.spark.writers.delta.utils import SparkConnectDeltaTableException

pytestmark = pytest.mark.spark


def test_scd2_custom_logic(spark):
    from koheesio.spark.utils.connect import is_remote_session

    def _get_result(target_df: DataFrame, expr: str):
        res = (
            target_df.where(expr)
            .select(
                "id",
                "last_updated_at",
                "meta.valid_from_timestamp",
                "meta.valid_to_timestamp",
                "value",
            )
            .orderBy("id", "valid_from_timestamp")
            .collect()[0]
            .asDict()
        )

        return res

    class SpecialSCD2DeltaTableWriter(SCD2DeltaTableWriter):
        orphaned_records_close: bool = Field(default=False, description="Close orphaned records")
        orphaned_records_close_ts: Optional[Column] = Field(
            default=None, description="Close orphaned records timestamp.Default to current timestamp UTC"
        )

        @staticmethod
        def _scd2_effective_time(meta_scd2_effective_time_col: str, **_kwargs) -> Column:
            scd2_effective_time = F.expr(
                f"""CASE 
                        WHEN __meta_scd2_system_merge_action='UC' AND cross.__meta_scd2_rn=1 THEN __meta_scd2_timestamp
                        WHEN __meta_scd2_system_merge_action='I' THEN TIMESTAMP'1970-01-01'
                        ELSE COALESCE({meta_scd2_effective_time_col},__meta_scd2_timestamp)
                    END 
                """
            )

            return scd2_effective_time

        @staticmethod
        def _scd2_end_time(meta_scd2_end_time_col: str, **_kwargs) -> Column:
            scd2_end_time = F.expr(
                f"""CASE 
                    WHEN __meta_scd2_system_merge_action='UC' AND cross.__meta_scd2_rn=2 THEN __meta_scd2_timestamp
                    WHEN __meta_scd2_system_merge_action='I' THEN TIMESTAMP'2999-12-31'
                    ELSE tgt.{meta_scd2_end_time_col} END"""
            )

            return scd2_end_time

        def _prepare_merge_builder(
            self,
            delta_table: DeltaTable,
            dest_alias: str,
            staged: DataFrame,
            merge_key: str,
            columns_to_process: List[str],
            meta_scd2_effective_time_col: str,
            **_kwargs,
        ) -> DeltaMergeBuilder:
            merge_builder = super()._prepare_merge_builder(
                delta_table=delta_table,
                dest_alias=dest_alias,
                staged=staged,
                merge_key=merge_key,
                columns_to_process=columns_to_process,
                meta_scd2_effective_time_col=meta_scd2_effective_time_col,
            )

            if self.orphaned_records_close:
                if self.orphaned_records_close_ts is None:
                    self.orphaned_records_close_ts = current_timestamp_utc(spark=self.spark)

                merge_builder = merge_builder.whenNotMatchedBySourceUpdate(
                    condition=F.col(
                        f"{self.meta_scd2_struct_col_name}.{self.meta_scd2_is_current_col_name}"
                    ).eqNullSafe(F.lit(True)),
                    set={
                        f"{self.meta_scd2_struct_col_name}.{self.meta_scd2_end_time_col_name}": self.orphaned_records_close_ts,
                    },
                )  # type: ignore

            return merge_builder

    target_table = "test_deltatable_writer_scd2"

    spark.sql(
        """CREATE OR REPLACE TABLE test_deltatable_writer_scd2 (
            id int NOT NULL,
            value STRING,
            last_updated_at TIMESTAMP,
            snapshot_at TIMESTAMP,
            meta STRUCT<valid_from_timestamp: TIMESTAMP, valid_to_timestamp: TIMESTAMP, is_current: BOOLEAN>
            )
            USING delta"""
    )

    source_df = spark.createDataFrame(
        [
            {"id": 2, "value": "value-2", "last_updated_at": "2024-02-01"},
            {"id": 3, "value": None, "last_updated_at": "2024-03-01"},
            {"id": 4, "value": "value-4", "last_updated_at": "2024-04-01"},
            {"id": 5, "value": "value-5", "last_updated_at": "2024-05-01"},
        ]
    )

    source_df = source_df.withColumn("last_updated_at", F.expr("to_timestamp(last_updated_at)")).withColumn(
        "snapshot_at", F.expr("TIMESTAMP'2024-12-31'")
    )

    writer = SpecialSCD2DeltaTableWriter(
        table=DeltaTableStep(table=target_table),
        merge_key="id",
        scd2_columns=["value"],
        scd2_timestamp_col=F.col("last_updated_at"),
        meta_scd2_struct_col_name="meta",
        meta_scd2_effective_time_col_name="valid_from_timestamp",
        meta_scd2_end_time_col_name="valid_to_timestamp",
        df=source_df,
    )

    if 3.4 < SPARK_MINOR_VERSION < 4.0 and is_remote_session():
        with pytest.raises(SparkConnectDeltaTableException) as exc_info:
            writer.execute()

        assert str(exc_info.value).startswith("`DeltaTable.forName` is not supported due to delta calling _sc")
    else:
        writer.execute()

        expected = {
            "id": 4,
            "last_updated_at": datetime.datetime(2024, 4, 1, 0, 0),
            "valid_from_timestamp": datetime.datetime(1970, 1, 1, 0, 0),
            "valid_to_timestamp": datetime.datetime(2999, 12, 31, 0, 0),
            "value": "value-4",
        }

        target_df = spark.read.table(target_table)
        result = _get_result(target_df, "id = 4")

        assert spark.table(target_table).count() == 4
        assert spark.table(target_table).where("meta.valid_to_timestamp = '2999-12-31'").count() == 4
        assert result == expected

    source_df2 = source_df.withColumn(
        "value", F.expr("CASE WHEN id = 2 THEN 'value-2-change' ELSE value END")
    ).withColumn("last_updated_at", F.expr("CASE WHEN id = 2 THEN TIMESTAMP'2024-02-02' ELSE last_updated_at END"))

    writer.df = source_df2

    if 3.4 < SPARK_MINOR_VERSION < 4.0 and is_remote_session():
        with pytest.raises(SparkConnectDeltaTableException) as exc_info:
            writer.execute()

        assert str(exc_info.value).startswith("`DeltaTable.forName` is not supported due to delta calling _sc")
    else:
        writer.execute()

        expected_insert = {
            "id": 2,
            "last_updated_at": datetime.datetime(2024, 2, 2, 0, 0),
            "valid_from_timestamp": datetime.datetime(2024, 2, 2, 0, 0),
            "valid_to_timestamp": datetime.datetime(2999, 12, 31, 0, 0),
            "value": "value-2-change",
        }

        expected_update = {
            "id": 2,
            "last_updated_at": datetime.datetime(2024, 2, 1, 0, 0),
            "valid_from_timestamp": datetime.datetime(1970, 1, 1, 0, 0),
            "valid_to_timestamp": datetime.datetime(2024, 2, 2, 0, 0),
            "value": "value-2",
        }

        result_insert = _get_result(target_df, "id = 2 and meta.valid_to_timestamp = '2999-12-31'")
        result_update = _get_result(target_df, "id = 2 and meta.valid_from_timestamp = '1970-01-01'")

        assert spark.table(target_table).count() == 5
        assert spark.table(target_table).where("meta.valid_to_timestamp = '2999-12-31'").count() == 4
        assert result_insert == expected_insert
        assert result_update == expected_update

    source_df3 = source_df2.withColumn(
        "value", F.expr("CASE WHEN id = 3 THEN 'value-3-change' ELSE value END")
    ).withColumn("last_updated_at", F.expr("CASE WHEN id = 3 THEN TIMESTAMP'2024-03-02' ELSE last_updated_at END"))

    writer.df = source_df3

    if 3.4 < SPARK_MINOR_VERSION < 4.0 and is_remote_session():
        with pytest.raises(SparkConnectDeltaTableException) as exc_info:
            writer.execute()

        assert str(exc_info.value).startswith("`DeltaTable.forName` is not supported due to delta calling _sc")
    else:
        writer.execute()

        expected_insert = {
            "id": 3,
            "last_updated_at": datetime.datetime(2024, 3, 2, 0, 0),
            "valid_from_timestamp": datetime.datetime(2024, 3, 2, 0, 0),
            "valid_to_timestamp": datetime.datetime(2999, 12, 31, 0, 0),
            "value": "value-3-change",
        }

        expected_update = {
            "id": 3,
            "last_updated_at": datetime.datetime(2024, 3, 1, 0, 0),
            "valid_from_timestamp": datetime.datetime(1970, 1, 1, 0, 0),
            "valid_to_timestamp": datetime.datetime(2024, 3, 2, 0, 0),
            "value": None,
        }

        result_insert = _get_result(target_df, "id = 3 and meta.valid_to_timestamp = '2999-12-31'")
        result_update = _get_result(target_df, "id = 3 and meta.valid_from_timestamp = '1970-01-01'")

        assert spark.table(target_table).count() == 6
        assert spark.table(target_table).where("meta.valid_to_timestamp = '2999-12-31'").count() == 4
        assert result_insert == expected_insert
        assert result_update == expected_update

    source_df4 = source_df3.where("id != 4")

    writer.df = source_df4
    if 3.4 < SPARK_MINOR_VERSION < 4.0 and is_remote_session():
        with pytest.raises(SparkConnectDeltaTableException) as exc_info:
            writer.execute()

        assert str(exc_info.value).startswith("`DeltaTable.forName` is not supported due to delta calling _sc")
    else:
        writer.execute()

        assert spark.table(target_table).count() == 6
        assert spark.table(target_table).where("id = 4 and meta.valid_to_timestamp = '2999-12-31'").count() == 1

    writer.orphaned_records_close = True
    if 3.4 < SPARK_MINOR_VERSION < 4.0 and is_remote_session():
        with pytest.raises(SparkConnectDeltaTableException) as exc_info:
            writer.execute()

        assert str(exc_info.value).startswith("`DeltaTable.forName` is not supported due to delta calling _sc")
    else:
        writer.execute()

        assert spark.table(target_table).count() == 6
        assert spark.table(target_table).where("id = 4 and meta.valid_to_timestamp = '2999-12-31'").count() == 0

    source_df5 = source_df4.where("id != 5")
    writer.orphaned_records_close_ts = F.col("snapshot_at")
    writer.df = source_df5

    if 3.4 < SPARK_MINOR_VERSION < 4.0 and is_remote_session():
        with pytest.raises(SparkConnectDeltaTableException) as exc_info:
            writer.execute()

        assert str(exc_info.value).startswith("`DeltaTable.forName` is not supported due to delta calling _sc")
    else:
        writer.execute()

        expected = {
            "id": 5,
            "last_updated_at": datetime.datetime(2024, 5, 1, 0, 0),
            "valid_from_timestamp": datetime.datetime(1970, 1, 1, 0, 0),
            "valid_to_timestamp": datetime.datetime(2024, 12, 31, 0, 0),
            "value": "value-5",
        }
        result = _get_result(target_df, "id = 5")

        assert spark.table(target_table).count() == 6
        assert spark.table(target_table).where("id = 5 and meta.valid_to_timestamp = '2999-12-31'").count() == 0
        assert result == expected


def test_scd2_logic(spark):
    from koheesio.spark.utils.connect import is_remote_session

    changes_data = [
        [("key1", "value1", "scd1-value11", "2024-05-01"), ("key2", "value2", "scd1-value21", "2024-04-01")],
        [("key1", "value1_updated", "scd1-value12", "2024-05-02"), ("key3", "value3", "scd1-value31", "2024-05-03")],
        [
            ("key1", "value1_updated_again", "scd1-value13", "2024-05-15"),
            ("key2", "value2_updated", "scd1-value22", "2024-05-14"),
        ],
    ]

    expected_data = [
        [
            Row(
                merge_key="key1",
                value_scd2="value1",
                value_scd1="scd1-value11",
                effective_time=datetime.datetime(2024, 5, 1, 0, 0),
                end_time=None,
                is_current=True,
            ),
            Row(
                merge_key="key2",
                value_scd2="value2",
                value_scd1="scd1-value21",
                effective_time=datetime.datetime(2024, 4, 1, 0, 0),
                end_time=None,
                is_current=True,
            ),
        ],
        [
            Row(
                merge_key="key1",
                value_scd2="value1",
                value_scd1="scd1-value11",
                effective_time=datetime.datetime(2024, 5, 1, 0, 0),
                end_time=datetime.datetime(2024, 5, 2, 0, 0),
                is_current=False,
            ),
            Row(
                merge_key="key1",
                value_scd2="value1_updated",
                value_scd1="scd1-value12",
                effective_time=datetime.datetime(2024, 5, 2, 0, 0),
                end_time=None,
                is_current=True,
            ),
            Row(
                merge_key="key2",
                value_scd2="value2",
                value_scd1="scd1-value21",
                effective_time=datetime.datetime(2024, 4, 1, 0, 0),
                end_time=None,
                is_current=True,
            ),
            Row(
                merge_key="key3",
                value_scd2="value3",
                value_scd1="scd1-value31",
                effective_time=datetime.datetime(2024, 5, 3, 0, 0),
                end_time=None,
                is_current=True,
            ),
        ],
        [
            Row(
                merge_key="key1",
                value_scd2="value1",
                value_scd1="scd1-value11",
                effective_time=datetime.datetime(2024, 5, 1, 0, 0),
                end_time=datetime.datetime(2024, 5, 2, 0, 0),
                is_current=False,
            ),
            Row(
                merge_key="key1",
                value_scd2="value1_updated",
                value_scd1="scd1-value12",
                effective_time=datetime.datetime(2024, 5, 2, 0, 0),
                end_time=datetime.datetime(2024, 5, 15, 0, 0),
                is_current=False,
            ),
            Row(
                merge_key="key1",
                value_scd2="value1_updated_again",
                value_scd1="scd1-value13",
                effective_time=datetime.datetime(2024, 5, 15, 0, 0),
                end_time=None,
                is_current=True,
            ),
            Row(
                merge_key="key2",
                value_scd2="value2",
                value_scd1="scd1-value21",
                effective_time=datetime.datetime(2024, 4, 1, 0, 0),
                end_time=datetime.datetime(2024, 5, 14, 0, 0),
                is_current=False,
            ),
            Row(
                merge_key="key2",
                value_scd2="value2_updated",
                value_scd1="scd1-value22",
                effective_time=datetime.datetime(2024, 5, 14, 0, 0),
                end_time=None,
                is_current=True,
            ),
            Row(
                merge_key="key3",
                value_scd2="value3",
                value_scd1="scd1-value31",
                effective_time=datetime.datetime(2024, 5, 3, 0, 0),
                end_time=None,
                is_current=True,
            ),
        ],
    ]
    spark.sql(
        """CREATE OR REPLACE TABLE scd2_test_data_set (
                merge_key STRING NOT NULL,
                value_scd2 STRING NOT NULL,
                value_scd1 STRING NOT NULL,
                _scd2 STRUCT<effective_time: TIMESTAMP, end_time: TIMESTAMP, is_current: BOOLEAN>
                )
                USING delta"""
    )

    writer = SCD2DeltaTableWriter(
        table=DeltaTableStep(table="scd2_test_data_set"),
        scd2_timestamp_col=F.col("run_date"),
        exclude_columns=["run_date"],
        merge_key="merge_key",
        scd2_columns=["value_scd2"],
        scd1_columns=["value_scd1"],
    )

    # Act & Assert
    for changes, expected in zip(changes_data, expected_data):
        changes_df = spark.createDataFrame(changes, ["merge_key", "value_scd2", "value_scd1", "run_date"])
        changes_df = changes_df.withColumn("run_date", F.to_timestamp("run_date"))
        writer.df = changes_df
        if 3.4 < SPARK_MINOR_VERSION < 4.0 and is_remote_session():
            with pytest.raises(SparkConnectDeltaTableException) as exc_info:
                writer.execute()

            assert str(exc_info.value).startswith("`DeltaTable.forName` is not supported due to delta calling _sc")
        else:
            writer.execute()
            res = (
                spark.sql("SELECT merge_key,value_scd2, value_scd1, _scd2.* FROM scd2_test_data_set")
                .orderBy("merge_key", "effective_time")
                .collect()
            )

            assert res == expected



from typing import List, Type

import pytest
from requests import HTTPError
from requests.exceptions import RetryError
import responses

from pydantic import ValidationError

from koheesio.logger import LoggingFactory
from koheesio.models import SecretStr
from koheesio.steps.http import (
    HttpDeleteStep,
    HttpGetStep,
    HttpMethod,
    HttpPostStep,
    HttpPutStep,
    HttpStep,
)

BASE_URL = "https://42.koheesio.test"
GET_ENDPOINT = f"{BASE_URL}/get"
POST_ENDPOINT = f"{BASE_URL}/post"
PUT_ENDPOINT = f"{BASE_URL}/put"
DELETE_ENDPOINT = f"{BASE_URL}/delete"
STATUS_404_ENDPOINT = f"{BASE_URL}/status/404"
STATUS_500_ENDPOINT = f"{BASE_URL}/status/500"
STATUS_503_ENDPOINT = f"{BASE_URL}/status/503"

log = LoggingFactory.get_logger(name="test_http", inherit_from_koheesio=True)


@responses.activate
@pytest.mark.parametrize(
    "endpoint,step,method,return_value,expected_status_code",
    [
        pytest.param(
            GET_ENDPOINT,
            HttpGetStep,
            "get",
            {"job_id": "1234"},
            200,
            id="httpGetStep_success",
        ),
        pytest.param(GET_ENDPOINT, HttpGetStep, "get", {"message": "Error Code"}, 400, id="httpGetStep_error"),
        pytest.param(GET_ENDPOINT, HttpGetStep, "get", None, 404, id="httpGetStep_emptyresponse"),
        pytest.param(
            POST_ENDPOINT,
            HttpPostStep,
            "post",
            {"job_id": "1234"},
            200,
            id="httpPostStep_success",
        ),
        pytest.param(POST_ENDPOINT, HttpPostStep, "post", {"message": "Error Code"}, 400, id="httpPosStep_error"),
        pytest.param(
            PUT_ENDPOINT,
            HttpPutStep,
            "put",
            {"detail": "Updated"},
            202,
            id="httpPutStep_success",
        ),
        pytest.param(PUT_ENDPOINT, HttpPutStep, "put", {"message": "Error Code"}, 400, id="httpPutStep_error"),
        pytest.param(
            PUT_ENDPOINT, HttpPutStep, "put", {"message": "Validation Error"}, 422, id="httpPutStep_validationError"
        ),
        pytest.param(
            DELETE_ENDPOINT,
            HttpDeleteStep,
            "delete",
            {"detail": "Deleted"},
            204,
            id="httpDeleteStep_success",
        ),
        pytest.param(
            DELETE_ENDPOINT, HttpDeleteStep, "delete", {"message": "Error Code"}, 400, id="httpDeleteStep_error"
        ),
        pytest.param(
            DELETE_ENDPOINT,
            HttpDeleteStep,
            "delete",
            {"message": "Validation Error"},
            422,
            id="httpDeleteStep_validationError",
        ),
    ],
)
def test_http_step(endpoint: str, step: HttpStep, method: str, return_value: dict, expected_status_code: int) -> None:
    """Test basic HTTP methods with success and error responses"""
    responses.add(getattr(responses, method.upper()), endpoint, json=return_value, status=expected_status_code)

    step = step(url=endpoint)

    if expected_status_code not in (200, 202, 204):
        with pytest.raises(HTTPError) as excinfo:
            step.execute()
        assert excinfo.value.response.status_code == expected_status_code
    else:
        response = step.execute()
        assert response.status_code == expected_status_code
        assert response.response_json == return_value


@responses.activate
def test_http_step_with_valid_http_method() -> None:
    """Unit Testing the GET functions with valid method."""
    responses.add(responses.GET, GET_ENDPOINT, status=200)
    step = HttpStep(method="get", url=GET_ENDPOINT)
    step.execute()
    assert step.output.status_code == 200  # type: ignore


@responses.activate
def test_http_step_with_invalid_http_method():
    responses.add(responses.GET, GET_ENDPOINT, status=200)
    # Will be raised during class instantiation
    with pytest.raises(AttributeError):
        HttpStep(method="foo", url=GET_ENDPOINT).execute()


@responses.activate
def test_http_step_request():
    responses.add(responses.PUT, PUT_ENDPOINT, status=200)
    # The default method for HttpStep class is GET, however the method specified in `request` options is PUT and
    # it will override the default
    step = HttpStep(url=PUT_ENDPOINT)
    with step._request(method=HttpMethod.PUT) as response:
        assert response.status_code == 200


EXAMPLE_DIGEST_AUTH = (
    'Digest username="Zaphod", realm="galaxy@heartofgold.com", nonce="42answer", uri="/dir/restaurant.html", '
    'response="dontpanic42", opaque="babelfish"'
)


@responses.activate
@pytest.mark.parametrize(
    "params, expected",
    [
        pytest.param(
            dict(url=GET_ENDPOINT, headers={"Authorization": "Bearer token", "Content-Type": "application/json"}),
            "Bearer token",
            id="bearer_token_in_headers",
        ),
        pytest.param(
            dict(url=GET_ENDPOINT, headers={"Content-Type": "application/json"}, auth_header="Bearer token"),
            "Bearer token",
            id="bearer_token_through_auth_header",
        ),
        pytest.param(
            dict(url=GET_ENDPOINT, auth_header=EXAMPLE_DIGEST_AUTH),
            EXAMPLE_DIGEST_AUTH,
            id="digest_auth_with_nonce",
        ),
    ],
)
def test_get_headers(params: dict, expected: str, caplog: pytest.LogCaptureFixture) -> None:
    """
    Authorization headers are being converted into SecretStr under the hood to avoid dumping any
    sensitive content into logs. However, when calling the `get_headers` method, the SecretStr is being
    converted back to string, otherwise sensitive info would have looked like '**********'.
    """
    # Arrange and Act
    responses.add(responses.GET, params["url"], status=200)  # Mock the request to be always successful
    step = HttpStep(**params)
    caplog.set_level("DEBUG", logger=step.log.name)
    auth = step.headers.get("Authorization")
    step.execute()

    # Check that the token doesn't accidentally leak in the logs
    assert len(caplog.records) > 1, "No logs were generated"
    for record in caplog.records:
        assert expected not in record.message

    # Ensure that the Authorization header is properly parsed to a SecretStr
    assert auth is not None, "Authorization header is missing"
    assert isinstance(auth, SecretStr)
    assert str(auth) == "**********"
    assert auth.get_secret_value() == expected

    # Ensure that the Content-Type header is properly parsed while not being a SecretStr
    assert step.headers["Content-Type"] == "application/json"


@responses.activate
@pytest.mark.parametrize(
    "max_retries, endpoint, status_code, expected_count, error_type",
    [
        pytest.param(1, STATUS_503_ENDPOINT, 503, 2, RetryError, id="max_retries_2_503"),
        pytest.param(5, STATUS_503_ENDPOINT, 502, 6, RetryError, id="max_retries_6_502"),
        # Only 502, 503, 504 are included in the retry list
        pytest.param(3, STATUS_404_ENDPOINT, 404, 1, HTTPError, id="max_retries_1_404"),
    ],
)
def test_max_retries(
    max_retries: int, endpoint: str, status_code: int, expected_count: int, error_type: Type[Exception]
) -> None:
    """Test retry behavior for different status codes"""
    request_count = 0

    def request_callback(request):
        nonlocal request_count
        request_count += 1
        return (status_code, {}, "Error")

    responses.add_callback(responses.GET, endpoint, callback=request_callback)

    step = HttpGetStep(url=endpoint, max_retries=max_retries)

    with pytest.raises(error_type):
        step.execute()

    assert request_count == expected_count



from pathlib import Path

import pytest

from koheesio.spark import DataFrame, SparkSession  # type: ignore
from koheesio.spark.transformations.download_files import DownloadFileFromUrlTransformation  # type: ignore


@pytest.fixture
def input_df(spark: SparkSession) -> DataFrame:
    """A simple DataFrame containing two URLs."""
    return spark.createDataFrame(
        [
            (101, "http://www.textfiles.com/100/adventur.txt"),
            (102, "http://www.textfiles.com/100/arttext.fun"),
        ],
        ["key", "url"],
    )


@pytest.fixture
def download_path(tmp_path: Path) -> Path:
    _path = tmp_path / "downloads"
    _path.mkdir(exist_ok=True)
    return _path


class TestDownloadFileFromUrlTransformation:
    """
    Input DataFrame:

    | key | url                                        |
    |-----|--------------------------------------------|
    | 101 |	http://www.textfiles.com/100/adventur.txt  |
    | 102 |	http://www.textfiles.com/100/arttext.fun   |

    Output DataFrame:

    | key | url                                        | downloaded_file_path  |
    |-----|--------------------------------------------|-----------------------|
    | 101 |	http://www.textfiles.com/100/adventur.txt  | downloads/adventur.txt|
    | 102 |	http://www.textfiles.com/100/arttext.fun   | downloads/arttext.fun |

    """

    def test_downloading_files(self, input_df: DataFrame, download_path: Path) -> None:
        """Test that the files are downloaded and the DataFrame is transformed correctly."""
        # Arrange
        expected_data = [
            "downloads/adventur.txt",
            "downloads/arttext.fun",
        ]

        # Act
        transformed_df = DownloadFileFromUrlTransformation(
            column="url",
            download_path=download_path,
            target_column="downloaded_file_path",
        ).transform(input_df)
        actual_data = sorted(
            [row.asDict()["downloaded_file_path"] for row in transformed_df.select("downloaded_file_path").collect()]
        )

        # Assert

        # Check that adventur.txt and arttext.fun are actually downloaded
        assert (download_path / "adventur.txt").exists()
        assert (download_path / "arttext.fun").exists()

        assert transformed_df.count() == 2
        assert transformed_df.columns == ["key", "url", "downloaded_file_path"]

        # check that the rows of the output DataFrame are as expected
        assert actual_data == expected_data



"""
The Context module is a part of the Koheesio framework and is primarily used for managing the environment configuration
where a Task or Step runs. It helps in adapting the behavior of a Task/Step based on the environment it operates in,
thereby avoiding the repetition of configuration values across different tasks.

The Context class, which is a key component of this module, functions similarly to a dictionary but with additional
features. It supports operations like handling nested keys, recursive merging of contexts, and
serialization/deserialization to and from various formats like JSON, YAML, and TOML.

For a comprehensive guide on the usage, examples, and additional features of the Context class, please refer to the
[reference/concepts/context](../reference/concepts/context.md) section of the Koheesio documentation.
"""

from __future__ import annotations

from typing import Any, Dict, Iterator, Union
from collections.abc import Mapping
from pathlib import Path
import re

import jsonpickle  # type: ignore[import-untyped]
import tomli
import yaml

__all__ = ["Context"]


class Context(Mapping):
    """
    The Context class is a key component of the Koheesio framework, designed to manage configuration data and shared
    variables across tasks and steps in your application. It behaves much like a dictionary, but with added
    functionalities.

    Key Features
    ------------
    - _Nested keys_: Supports accessing and adding nested keys similar to dictionary keys.
    - _Recursive merging_: Merges two Contexts together, with the incoming Context having priority.
    - _Serialization/Deserialization_: Easily created from a yaml, toml, or json file, or a dictionary, and can be
        converted back to a dictionary.
    - _Handling complex Python objects_: Uses jsonpickle for serialization and deserialization of complex Python objects
        to and from JSON.

    For a comprehensive guide on the usage, examples, and additional features of the Context class, please refer to the
    [reference/concepts/context](../reference/concepts/context.md) section of the Koheesio documentation.

    Methods
    -------
    add(key: str, value: Any) -> Context
        Add a key/value pair to the context.
    get(key: str, default: Any = None, safe: bool = True) -> Any`
        Get value of a given key.
    get_item(key: str, default: Any = None, safe: bool = True) -> Dict[str, Any]
        Acts just like `.get`, except that it returns the key also.
    contains(key: str) -> bool
        Check if the context contains a given key.
    merge(context: Context, recursive: bool = False) -> Context
        Merge this context with the context of another, where the incoming context has priority.
    to_dict() -> dict
        Returns all parameters of the context as a dict.
    from_dict(kwargs: dict) -> Context
        Creates Context object from the given dict.
    from_yaml(yaml_file: str | Path) -> Context
        Creates Context object from a given yaml file.
    from_json(json_file: str | Path) -> Context
        Creates Context object from a given json file.

    Dunder methods
    --------------

    - _`_iter__()`: Allows for iteration across a Context.
    - `__len__()`: Returns the length of the Context.
    - `__getitem__(item)`: Makes class subscriptable.

    Inherited from Mapping
    ----------------------

    - `items()`: Returns all items of the Context.
    - `keys()`: Returns all keys of the Context.
    - `values()`: Returns all values of the Context.
    """

    def __init__(self, *args, **kwargs):  # type: ignore[no-untyped-def]
        """Initializes the Context object with given arguments."""
        for arg in args:
            if isinstance(arg, dict):
                kwargs.update(arg)
            if isinstance(arg, Context):
                kwargs.update(arg.to_dict())

        if kwargs:
            for key, value in kwargs.items():
                self.__dict__[key] = self.process_value(value)

    def __str__(self) -> str:
        """Returns a string representation of the Context."""
        return str(dict(self.__dict__))

    def __repr__(self) -> str:
        """Returns a string representation of the Context."""
        return self.__str__()

    def __iter__(self) -> Iterator[str]:
        """Allows for iteration across a Context"""
        return self.to_dict().__iter__()

    def __len__(self) -> int:
        """Returns the length of the Context"""
        return self.to_dict().__len__()

    def __getattr__(self, item: str) -> Any:
        try:
            return self.get(item, safe=False)
        except KeyError as e:
            raise AttributeError(item) from e

    def __getitem__(self, item: str) -> Any:
        """Makes class subscriptable"""
        return self.get(item, safe=False)

    @classmethod
    def _recursive_merge(cls, target_context: Context, merge_context: Context) -> Context:
        """
        Recursively merge two dictionaries to an arbitrary depth.

        Parameters
        ----------
        target_context: Context
            Koheesio Context to merge into
        merge_context: Context
            Koheesio Context that is being merged

        Returns
        -------
        Context

        Example
        --------
        ```python
        context_1 = {
            "key_1": "val_1",
            "key_2": {
                "sub_key_1": "sub_val_1",
                "sub_key_2": "sub_val_2",
            },
            "key_3": ["item_1"]
        }

        context_2 = {
            "key_2": {
                "sub_key_2": "sub_val_2.1",
                "sub_key_3": "sub_val_3",
            },
            "key_3": ["item_2"]
        }

        result = {
            "key_1": "val_1",
            "key_2": {
                "sub_key_1": "sub_val_1",
                "sub_key_2": "sub_val_2.1",
                "sub_key_3": "sub_val_3",
            }
            "key_3": ["item_1","item_2]
        }
        ```
        """
        for k, v in merge_context.items():
            if k in target_context and isinstance(target_context[k], Context) and isinstance(v, Context):
                cls._recursive_merge(target_context[k], merge_context[k])
            elif k in target_context and isinstance(target_context[k], list) and isinstance(v, list):
                # merge list items
                target_context[k].extend(merge_context[k])
            else:
                target_context.add(k, v)
        return target_context

    @classmethod
    def from_dict(cls, kwargs: dict) -> Context:
        """Creates Context object from the given dict

        Parameters
        ----------
        kwargs: dict

        Returns
        -------
        Context
        """
        return cls(kwargs)

    @classmethod
    def from_json(cls, json_file_or_str: Union[str, Path]) -> Context:
        """Creates Context object from a given json file

        Note: jsonpickle is used to serialize/deserialize the Context object. This is done to allow for objects to be
        stored in the Context object, which is not possible with the standard json library.

        Why jsonpickle?
        ---------------
        (from https://jsonpickle.github.io/)

        > Data serialized with python’s pickle (or cPickle or dill) is not easily readable outside of python. Using the
        json format, jsonpickle allows simple data types to be stored in a human-readable format, and more complex
        data types such as numpy arrays and pandas dataframes, to be machine-readable on any platform that supports
        json.

        Security
        --------
        (from https://jsonpickle.github.io/)

        > jsonpickle should be treated the same as the Python stdlib pickle module from a security perspective.

        ### ! Warning !
        > The jsonpickle module is not secure. Only unpickle data you trust.
        It is possible to construct malicious pickle data which will execute arbitrary code during unpickling.
        Never unpickle data that could have come from an untrusted source, or that could have been tampered with.
        Consider signing data with an HMAC if you need to ensure that it has not been tampered with.
        Safer deserialization approaches, such as reading JSON directly, may be more appropriate if you are processing
        untrusted data.

        Parameters
        ----------
        json_file_or_str : Union[str, Path]
            Pathlike string or Path that points to the json file or string containing json

        Returns
        -------
        Context
        """
        json_str = json_file_or_str

        # check if json_str is pathlike
        if (json_file := Path(json_file_or_str)).exists():
            json_str = json_file.read_text(encoding="utf-8")

        json_dict = jsonpickle.loads(json_str)
        return cls.from_dict(json_dict)

    @classmethod
    def from_toml(cls, toml_file_or_str: Union[str, Path]) -> Context:
        """Creates Context object from a given toml file

        Parameters
        ----------
        toml_file_or_str: Union[str, Path]
            Pathlike string or Path that points to the toml file or string containing toml

        Returns
        -------
        Context
        """

        # check if toml_str is pathlike
        if (toml_file := Path(toml_file_or_str)).exists():
            toml_str = toml_file.read_text(encoding="utf-8")
        else:
            toml_str = str(toml_file_or_str)

        toml_dict = tomli.loads(toml_str)
        return cls.from_dict(toml_dict)

    @classmethod
    def from_yaml(cls, yaml_file_or_str: str) -> Context:
        """Creates Context object from a given yaml file

        Parameters
        ----------
        yaml_file_or_str: str or Path
            Pathlike string or Path that points to the yaml file, or string containing yaml

        Returns
        -------
        Context
        """
        yaml_str = yaml_file_or_str

        # check if yaml_str is pathlike
        if (yaml_file := Path(yaml_file_or_str)).exists():
            yaml_str = yaml_file.read_text(encoding="utf-8")

        # Bandit: disable yaml.load warning
        yaml_dict = yaml.load(yaml_str, Loader=yaml.Loader)  # nosec B506: yaml_load

        return cls.from_dict(yaml_dict)

    def add(self, key: str, value: Any) -> Context:
        """Add a key/value pair to the context"""
        self.__dict__[key] = value
        return self

    def contains(self, key: str) -> bool:
        """Check if the context contains a given key

        Parameters
        ----------
        key: str

        Returns
        -------
        bool
        """
        try:
            self.get(key, safe=False)
            return True
        except KeyError:
            return False

    def get(self, key: str, default: Any = None, safe: bool = True) -> Any:
        """Get value of a given key

        The key can either be an actual key (top level) or the key of a nested value.
        Behaves a lot like a dict's `.get()` method otherwise.

        Parameters
        ----------
        key:
            Can be a real key, or can be a dotted notation of a nested key
        default:
            Default value to return
        safe:
            Toggles whether to fail or not when item cannot be found

        Returns
        -------
        Any
            Value of the requested item

        Example
        -------
        Example of a nested call:

        ```python
        context = Context({"a": {"b": "c", "d": "e"}, "f": "g"})
        context.get("a.b")
        ```

        Returns `c`
        """
        try:
            # in case key is directly available, or is written in dotted notation
            try:
                return self.__dict__[key]
            except KeyError:
                pass
            if "." in key:
                # handle nested keys
                nested_keys = key.split(".")
                value = self  # parent object
                for k in nested_keys:
                    value = value[k]  # iterate through nested values
                return value

            raise KeyError

        except (AttributeError, KeyError, TypeError) as e:
            if not safe:
                raise KeyError(f"requested key '{key}' does not exist in {self}") from e
            return default

    def get_item(self, key: str, default: Any = None, safe: bool = True) -> Dict[str, Any]:
        """Acts just like `.get`, except that it returns the key also

        Returns
        -------
        Dict[str, Any]
            key/value-pair of the requested item

        Example
        -------
        Example of a nested call:

        ```python
        context = Context({"a": {"b": "c", "d": "e"}, "f": "g"})
        context.get_item("a.b")
        ```

        Returns `{'a.b': 'c'}`
        """
        value = self.get(key, default, safe)
        return {key: value}

    def get_all(self) -> dict:
        """alias to to_dict()"""
        return self.to_dict()

    def merge(self, context: Context, recursive: bool = False) -> Context:
        """Merge this context with the context of another, where the incoming context has priority.

        Parameters
        ----------
        context: Context
            Another Context class
        recursive: bool
            Recursively merge two dictionaries to an arbitrary depth

        Returns
        -------
        Context
            updated context
        """
        if recursive:
            return Context.from_dict(self._recursive_merge(target_context=self, merge_context=context).to_dict())

        # just merge on the top level keys
        return Context.from_dict({**self.to_dict(), **context.to_dict()})

    def process_value(self, value: Any) -> Any:
        """Processes the given value, converting dictionaries to Context objects as needed."""
        if isinstance(value, dict):
            return self.from_dict(value)

        if isinstance(value, (list, set)):
            return [self.from_dict(v) if isinstance(v, dict) else v for v in value]

        return value

    def to_dict(self) -> Dict[str, Any]:
        """Returns all parameters of the context as a dict

        Returns
        -------
        dict
            containing all parameters of the context
        """
        result = {}

        for key, value in self.__dict__.items():
            if isinstance(value, Context):
                result[key] = value.to_dict()
            elif isinstance(value, list):
                result[key] = [e.to_dict() if isinstance(e, Context) else e for e in value]  # type: ignore[assignment]
            else:
                result[key] = value

        return result

    def to_json(self, pretty: bool = False) -> str:
        """Returns all parameters of the context as a json string

        Note: jsonpickle is used to serialize/deserialize the Context object. This is done to allow for objects to be
        stored in the Context object, which is not possible with the standard json library.

        Why jsonpickle?
        ---------------
        (from https://jsonpickle.github.io/)

        > Data serialized with python's pickle (or cPickle or dill) is not easily readable outside of python. Using the
        json format, jsonpickle allows simple data types to be stored in a human-readable format, and more complex
        data types such as numpy arrays and pandas dataframes, to be machine-readable on any platform that supports
        json.

        Parameters
        ----------
        pretty : bool, optional, default=False
            Toggles whether to return a pretty json string or not

        Returns
        -------
        str
            containing all parameters of the context
        """
        d = self.to_dict()
        return jsonpickle.dumps(d, indent=4) if pretty else jsonpickle.dumps(d)

    def to_yaml(self, clean: bool = False) -> str:
        """Returns all parameters of the context as a yaml string

        Parameters
        ----------
        clean: bool
            Toggles whether to remove `!!python/object:...` from yaml or not.
            Default: False

        Returns
        -------
        str
            containing all parameters of the context
        """
        # sort_keys=False to preserve order of keys
        yaml_str = yaml.dump(self.to_dict(), sort_keys=False)

        # remove `!!python/object:...` from yaml
        if clean:
            remove_pattern = re.compile(r"!!python/object:.*?\n")
            yaml_str = re.sub(remove_pattern, "\n", yaml_str)

        return yaml_str



import pytest

from koheesio.spark.readers.spark_sql_reader import SparkSqlReader

pytestmark = pytest.mark.spark


def test_spark_sql_reader(spark, data_path):
    spark.sql("CREATE TABLE IF NOT EXISTS table_A (id INT, a_name STRING) USING DELTA")
    spark.sql("CREATE TABLE IF NOT EXISTS table_B (id INT, b_name STRING) USING DELTA")
    spark.sql("INSERT INTO table_A VALUES (1, 'Table_A'), (2, 'B'), (3, 'C')")
    spark.sql("INSERT INTO table_B VALUES (1, 'Table_B'), (2, 'D'), (3, 'E')")

    reader = SparkSqlReader(
        sql="""SELECT * FROM table_A JOIN table_B ON table_A.${join_col} = table_B.${join_col}
               WHERE table_A.${join_col} = 1""",
        join_col="id",
    )
    df = reader.read()
    actual = df.head().asDict()
    expected = {"id": 1, "a_name": "Table_A", "b_name": "Table_B"}
    assert actual == expected

    reader = SparkSqlReader(sql_path=f"{data_path}/sql/spark_sql_reader.sql")
    df = reader.read()
    actual = df.head().asDict()
    assert actual == expected


def test_spark_sql_reader_failed():
    with pytest.raises(ValueError):
        SparkSqlReader(sql="SELECT 1", sql_path="tests/resources/sql/none_existent_path.sql")

    with pytest.raises(FileNotFoundError):
        SparkSqlReader(sql_path="tests/resources/sql/none_existent_path.sql")

    with pytest.raises(ValueError):
        SparkSqlReader(sql=None)

    with pytest.raises(ValueError):
        SparkSqlReader(sql="")



"""
This module is the entry point for the koheesio.spark.writers.delta package.

It imports and exposes the DeltaTableWriter and DeltaTableStreamWriter classes for external use.

Classes:
    DeltaTableWriter: Class to write data in batch mode to a Delta table.
    DeltaTableStreamWriter: Class to write data in streaming mode to a Delta table.
"""

from koheesio.spark.writers.delta.batch import BatchOutputMode, DeltaTableWriter
from koheesio.spark.writers.delta.scd import SCD2DeltaTableWriter
from koheesio.spark.writers.delta.stream import DeltaTableStreamWriter

__all__ = ["DeltaTableWriter", "DeltaTableStreamWriter", "SCD2DeltaTableWriter", "BatchOutputMode"]



from typing import Any
from pathlib import Path

import pytest
from requests_mock.mocker import Mocker

from koheesio import LoggingFactory  # type: ignore
from koheesio.steps.download_file import DownloadFileStep, FileWriteMode  # type: ignore

URL = "http://example.com/testfile.txt"
log = LoggingFactory.get_logger(name="test_download_file", inherit_from_koheesio=True)


@pytest.fixture
def large_file_path(data_path: str) -> Path:
    return Path(f"{data_path}/transformations/string_data/100000_rows_with_strings.json")


@pytest.fixture
def download_path(tmp_path: Path) -> Path:
    _path = tmp_path / "downloads"
    _path.mkdir(exist_ok=True)
    return _path


@pytest.fixture
def downloaded_file(download_path: Path) -> Path:
    _file = download_path / "testfile.txt"
    _file.touch(exist_ok=True)
    return _file


class TestFileWriteMode:
    """Test the FileWriteMode enum"""

    @pytest.mark.parametrize(
        "mode, expected",
        [
            (FileWriteMode.OVERWRITE, "wb"),
            (FileWriteMode.BACKUP, "wb"),
            (FileWriteMode.EXCLUSIVE, "wb"),
            (FileWriteMode.APPEND, "ab"),
        ],
    )
    def test_write_mode(self, mode: FileWriteMode, expected: str) -> None:
        """Test that the write mode is correctly set for each FileWriteMode"""
        assert mode.write_mode == expected

    @pytest.mark.parametrize(
        "mode_str, expected_mode",
        [
            ("overwrite", FileWriteMode.OVERWRITE),
            ("append", FileWriteMode.APPEND),
            ("ignore", FileWriteMode.IGNORE),
            ("exclusive", FileWriteMode.EXCLUSIVE),
            ("backup", FileWriteMode.BACKUP),
        ],
    )
    def test_from_string(self, mode_str: str, expected_mode: FileWriteMode) -> None:
        """Test that the FileWriteMode enum can be created from a string"""
        assert FileWriteMode.from_string(mode_str) == expected_mode


class TestDownloadFileStep:
    """Test the DownloadFileStep class"""

    def test_default_write_mode(self, download_path: Path) -> None:
        """Test that the default write mode is OVERWRITE"""
        step = DownloadFileStep(url=URL, download_path=download_path)
        assert step.mode == FileWriteMode.OVERWRITE

    def test_write_mode_conversion(self, download_path: Path) -> None:
        """Test that the write mode can be set using a string"""
        step = DownloadFileStep(url=URL, mode="append", download_path=download_path)
        assert step.mode == FileWriteMode.APPEND

    def test_large_file(self, download_path: Path, large_file_path: Path) -> None:
        """Test that a large file can be downloaded using default settings"""
        with Mocker() as mocker:
            # Arrange
            mocker.get(URL, content=large_file_path.read_bytes())

            # Act
            step = DownloadFileStep(url=URL, download_path=download_path)
            step.execute()

        # Assert
        downloaded_file = download_path / "testfile.txt"
        assert downloaded_file.exists()
        assert downloaded_file.read_bytes() == large_file_path.read_bytes()

    def test_invalid_write_mode(self) -> None:
        """Test that an error is raised for an invalid write mode"""
        with pytest.raises(ValueError):
            DownloadFileStep(
                mode="invalid_mode",
                url=URL,
            )

    def test_download_file_step_overwrite_mode(self, download_path: Path, downloaded_file: Any) -> None:
        """In OVERWRITE mode, the file should be overwritten if it exists"""
        # Arrange
        downloaded_file.write_bytes(b"foo")

        with Mocker() as mocker:
            mocker.get(URL, content=b"bar")

            # Act
            step = DownloadFileStep(url=URL, download_path=download_path, mode="overwrite")
            step.execute()

        # Assert
        assert downloaded_file.exists()
        assert downloaded_file.read_bytes() == b"bar"

    def test_download_file_step_append_mode(self, download_path: Path, downloaded_file: Any) -> None:
        """In APPEND mode, the file should be appended to if it exists"""
        # Arrange
        downloaded_file.write_bytes(b"foo")

        with Mocker() as mocker:
            mocker.get(URL, content=b"bar")

            # Act
            step = DownloadFileStep(url=URL, download_path=download_path, mode="append")
            step.execute()

        # Assert
        assert downloaded_file.exists()
        assert downloaded_file.read_bytes() == b"foobar"

    def test_download_file_step_ignore_mode(
        self, download_path: Path, downloaded_file: Any, caplog: pytest.LogCaptureFixture
    ) -> None:
        """In IGNORE mode, the class should return without writing anything if the file exists"""
        # Arrange
        downloaded_file.write_bytes(b"foo")

        with caplog.at_level("INFO"), Mocker() as mocker:
            mocker.get(URL, content=b"bar")

            # Act
            step = DownloadFileStep(url=URL, download_path=download_path, mode="ignore")
            step.log.setLevel("INFO")
            step.execute()

        # Assert
        assert "Ignoring testfile.txt based on IGNORE mode." in caplog.text
        assert downloaded_file.exists()
        assert downloaded_file.read_bytes() == b"foo"

    def test_download_file_step_exclusive_mode(self, download_path: Path, downloaded_file: Any) -> None:
        """In EXCLUSIVE mode, an error should be raised if the file exists"""
        # Arrange
        downloaded_file.write_bytes(b"foo")

        with Mocker() as mocker:
            mocker.get(URL, content=b"bar")

            # Act and Assert
            with pytest.raises(FileExistsError):
                step = DownloadFileStep(url=URL, download_path=download_path, mode="exclusive")
                step.execute()

    def test_download_file_step_backup_mode(self, download_path: Path, downloaded_file: Any) -> None:
        """In BACKUP mode, the file should be backed up before being overwritten"""
        # Arrange
        downloaded_file.write_bytes(b"foo")

        with Mocker() as mocker:
            mocker.get(URL, content=b"bar")

            # Act
            step = DownloadFileStep(url=URL, download_path=download_path, mode="backup")
            step.execute()

        # Assert
        assert downloaded_file.exists()
        assert downloaded_file.read_bytes() == "bar".encode()
        backup_files = list(download_path.glob("testfile.txt.*.bak"))
        assert len(backup_files) == 1
        assert backup_files[0].read_bytes() == "foo".encode()



"""
This module provides utility functions while working with delta framework.
"""

from typing import Optional

from delta import DeltaTable
from py4j.java_gateway import JavaObject

from koheesio.spark import SparkSession
from koheesio.spark.utils import SPARK_MINOR_VERSION


class SparkConnectDeltaTableException(AttributeError):
    EXCEPTION_CONNECT_TEXT: str = """`DeltaTable.forName` is not supported due to delta calling _sc, 
                    which is not available in Spark Connect and PySpark>=3.5,<4.0. Required version of PySpark >=4.0.
                    Possible workaround to use spark.read and Spark SQL for any Delta operation (e.g. merge)"""

    def __init__(self, original_exception: AttributeError):
        custom_message = f"{self.EXCEPTION_CONNECT_TEXT}\nOriginal exception: {str(original_exception)}"
        super().__init__(custom_message)


def log_clauses(clauses: JavaObject, source_alias: str, target_alias: str) -> Optional[str]:
    """
    Prepare log message for clauses of DeltaMergePlan statement.

    Parameters
    ----------
    clauses : JavaObject
        The clauses of the DeltaMergePlan statement.
    source_alias : str
        The source alias.
    target_alias : str
        The target alias.

    Returns
    -------
    Optional[str]
        The log message if there are clauses, otherwise None.

    Notes
    -----
    This function prepares a log message for the clauses of a DeltaMergePlan statement. It iterates over the clauses,
    processes the conditions, and constructs the log message based on the clause type and columns.

    If the condition is a value, it replaces the source and target aliases in the condition string. If the condition is
    None, it sets the condition_clause to "No conditions required".

    The log message includes the clauses type, the clause type, the columns, and the condition.
    """
    log_message = None

    if not clauses.isEmpty():
        clauses_type = clauses.last().nodeName().replace("DeltaMergeInto", "")
        _processed_clauses: dict = {}

        for i in range(0, clauses.length()):
            clause = clauses.apply(i)
            condition = clause.condition()

            if "value" in dir(condition):
                condition_clause = (
                    condition.value()
                    .toString()
                    .replace(f"'{source_alias}", source_alias)
                    .replace(f"'{target_alias}", target_alias)
                )
            elif condition.toString() == "None":
                condition_clause = "No conditions required"
            else:
                raise ValueError(f"Condition {condition} is not supported")

            clause_type: str = clause.clauseType().capitalize()
            columns = "ALL" if clause_type == "Delete" else clause.actions().toList().apply(0).toString()

            if clause_type.lower() not in _processed_clauses:
                _processed_clauses[clause_type.lower()] = []

            log_message = (
                f"{clauses_type} will perform action:{clause_type} columns ({columns}) if `{condition_clause}`"
            )

    return log_message


def get_delta_table_for_name(spark_session: SparkSession, table_name: str) -> DeltaTable:
    """
    Retrieves the DeltaTable instance for the specified table name.

    This method attempts to get the DeltaTable using the provided Spark session and table name.
    If an AttributeError occurs and the Spark version is between 3.4 and 4.0, and the session is remote,
    it raises a SparkConnectDeltaTableException.

    Parameters
    ----------
    spark_session : SparkSession
        The Spark Session to use.
    table_name : str
        The table name.

    Returns
    -------
    DeltaTable
        The DeltaTable instance for the specified table name.

    Raises
    ------
    SparkConnectDeltaTableException
        If the Spark version is between 3.4 and 4.0, the session is remote, and an AttributeError occurs.
    """
    try:
        delta_table = DeltaTable.forName(sparkSession=spark_session, tableOrViewName=table_name)
    except AttributeError as e:
        from koheesio.spark.utils.connect import is_remote_session

        if 3.4 < SPARK_MINOR_VERSION < 4.0 and is_remote_session():
            raise SparkConnectDeltaTableException(e) from e
        else:
            raise e

    return delta_table



"""
Module containing Snowflake reader classes.

This module contains classes for reading data from Snowflake. The classes are used to create a Spark DataFrame from a
Snowflake table or a query.

Classes
-------
SnowflakeReader
    Reader for Snowflake tables.
Query
    Reader for Snowflake queries.
DbTableQuery
    Reader for Snowflake queries that return a single row.

Notes
-----
The classes are defined in the koheesio.steps.integrations.snowflake module; this module simply inherits from the
classes defined there.

See Also
--------
- [koheesio.spark.readers.Reader](index.md#koheesio.spark.readers.Reader)
    Base class for all Readers.
- [koheesio.spark.snowflake](../snowflake.md)
    Module containing Snowflake classes.

More detailed class descriptions can be found in the class docstrings.
"""

from koheesio.spark.snowflake import DbTableQuery, Query, SnowflakeReader

__all__ = ["SnowflakeReader", "Query", "DbTableQuery"]



# flake8: noqa: F811
import logging
from textwrap import dedent
from unittest import mock
from unittest.mock import Mock

import pytest

from pyspark.sql import types as t

from koheesio.integrations.snowflake.test_utils import mock_query
from koheesio.integrations.spark.snowflake import (
    AddColumn,
    CreateOrReplaceTableFromDataFrame,
    DbTableQuery,
    GetTableSchema,
    Query,
    RunQuery,
    SnowflakeReader,
    SnowflakeWriter,
    SyncTableAndDataFrameSchema,
    TableExists,
    TagSnowflakeQuery,
    map_spark_type,
)
from koheesio.spark.writers import BatchOutputMode

pytestmark = pytest.mark.spark

COMMON_OPTIONS = {
    "url": "url",
    "user": "user",
    "password": "password",
    "database": "db",
    "schema": "schema",
    "role": "role",
    "warehouse": "warehouse",
}


def test_snowflake_module_import():
    # test that the pass-through imports in the koheesio.spark snowflake modules are working
    from koheesio.spark.readers import snowflake as snowflake_writers
    from koheesio.spark.writers import snowflake as snowflake_readers


class TestSnowflakeReader:
    reader_options = {"dbtable": "table", **COMMON_OPTIONS}

    def test_get_options(self):
        sf = SnowflakeReader(**(self.reader_options | {"authenticator": None}))
        o = sf.get_options()
        assert sf.format == "snowflake"
        assert o["sfUser"] == "user"
        assert o["sfCompress"] == "on"
        assert "authenticator" not in o

    def test_execute(self, dummy_spark):
        """Method should be callable from parent class"""
        k = SnowflakeReader(**self.reader_options).execute()
        assert k.df.count() == 3


class TestRunQuery:
    def test_deprecation(self):
        """Test for the deprecation warning"""
        with pytest.warns(
            DeprecationWarning, match="The RunQuery class is deprecated and will be removed in a future release."
        ):
            try:
                _ = RunQuery(
                    **COMMON_OPTIONS,
                    query="<deprecated>",
                )
            except RuntimeError:
                pass  # Ignore any RuntimeError that occur after the warning

    def test_spark_connect(self, spark):
        """Test that we get a RuntimeError when using a SparkSession without a JVM"""
        from koheesio.spark.utils.connect import is_remote_session

        if not is_remote_session(spark):
            pytest.skip(reason="Test only runs when we have a remote SparkSession")

        with pytest.raises(RuntimeError):
            _ = RunQuery(
                **COMMON_OPTIONS,
                query="<deprecated>",
            )


class TestQuery:
    options = {"query": "query", **COMMON_OPTIONS}

    def test_execute(self, dummy_spark):
        k = Query(**self.options).execute()
        assert k.df.count() == 3


class TestTableQuery:
    options = {"table": "table", **COMMON_OPTIONS}

    def test_execute(self, dummy_spark):
        k = DbTableQuery(**self.options).execute()
        assert k.df.count() == 3


class TestCreateOrReplaceTableFromDataFrame:
    options = {"table": "table", "account": "bar", **COMMON_OPTIONS}

    def test_execute(self, dummy_spark, dummy_df, mock_query):
        k = CreateOrReplaceTableFromDataFrame(**self.options, df=dummy_df).execute()
        assert k.snowflake_schema == "id BIGINT"
        assert k.query == "CREATE OR REPLACE TABLE db.schema.table (id BIGINT)"
        assert len(k.input_schema) > 0
        mock_query.assert_called_with(k.query)


class TestGetTableSchema:
    options = {"table": "table", **COMMON_OPTIONS}

    def test_execute(self, dummy_spark):
        k = GetTableSchema(**self.options)
        assert len(k.execute().table_schema.fields) == 2


class TestAddColumn:
    options = {"table": "foo", "column": "bar", "type": t.DateType(), "account": "foo", **COMMON_OPTIONS}

    def test_execute(self, dummy_spark, mock_query):
        k = AddColumn(**self.options).execute()
        assert k.query == "ALTER TABLE FOO ADD COLUMN BAR DATE"
        mock_query.assert_called_with(k.query)


class TestSnowflakeWriter:
    def test_execute(self, mock_df):
        k = SnowflakeWriter(
            **COMMON_OPTIONS,
            table="foo",
            df=mock_df,
            mode=BatchOutputMode.OVERWRITE,
        )
        k.execute()

        # check that the format was set to snowflake
        mocked_format: Mock = mock_df.write.format
        assert mocked_format.call_args[0][0] == "snowflake"
        mock_df.write.format.assert_called_with("snowflake")


class TestSyncTableAndDataFrameSchema:
    @mock.patch("koheesio.integrations.spark.snowflake.AddColumn")
    @mock.patch("koheesio.integrations.spark.snowflake.GetTableSchema")
    def test_execute(self, mock_get_table_schema, mock_add_column, spark, caplog):
        # Arrange
        from pyspark.sql.types import StringType, StructField, StructType

        df = spark.createDataFrame(data=[["val"]], schema=["foo"])
        sf_schema_before = StructType([StructField("bar", StringType(), True)])
        sf_schema_after = StructType([StructField("bar", StringType(), True), StructField("foo", StringType(), True)])

        mock_get_table_schema_instance = mock_get_table_schema()
        mock_get_table_schema_instance.execute.side_effect = [
            mock.Mock(table_schema=sf_schema_before),
            mock.Mock(table_schema=sf_schema_after),
        ]

        logger = logging.getLogger("koheesio")
        logger.setLevel(logging.WARNING)

        # Act and Assert -- dry run
        with caplog.at_level(logging.WARNING):
            k = SyncTableAndDataFrameSchema(
                **COMMON_OPTIONS,
                table="foo",
                df=df,
                dry_run=True,
            ).execute()
            assert "Columns to be added to Snowflake table: {'foo'}" in caplog.text
            assert "Columns to be added to Spark DataFrame: {'bar'}" in caplog.text
            assert k.new_df_schema == StructType()

        # Act and Assert -- execute
        k = SyncTableAndDataFrameSchema(
            **COMMON_OPTIONS,
            table="foo",
            df=df,
        ).execute()
        assert sorted(k.df.columns) == ["bar", "foo"]


@pytest.mark.parametrize(
    "input_value,expected",
    [
        (t.BinaryType(), "VARBINARY"),
        (t.BooleanType(), "BOOLEAN"),
        (t.ByteType(), "BINARY"),
        (t.DateType(), "DATE"),
        (t.TimestampType(), "TIMESTAMP"),
        (t.DoubleType(), "DOUBLE"),
        (t.FloatType(), "FLOAT"),
        (t.IntegerType(), "INT"),
        (t.LongType(), "BIGINT"),
        (t.NullType(), "STRING"),
        (t.ShortType(), "SMALLINT"),
        (t.StringType(), "STRING"),
        (t.NumericType(), "FLOAT"),
        (t.DecimalType(0, 1), "DECIMAL(0,1)"),
        (t.DecimalType(0, 100), "DECIMAL(0,100)"),
        (t.DecimalType(10, 0), "DECIMAL(10,0)"),
        (t.DecimalType(), "DECIMAL(10,0)"),
        (t.MapType(t.IntegerType(), t.StringType()), "VARIANT"),
        (t.ArrayType(t.StringType()), "VARIANT"),
        (t.StructType([t.StructField(name="foo", dataType=t.StringType())]), "VARIANT"),
        (t.DayTimeIntervalType(), "STRING"),
    ],
)
def test_map_spark_type(input_value, expected):
    assert map_spark_type(input_value) == expected


class TestTableExists:
    options = dict(
        sfURL="url",
        sfUser="user",
        sfPassword="password",
        sfDatabase="database",
        sfRole="role",
        sfWarehouse="warehouse",
        schema="schema",
        table="table",
    )

    def test_table_exists(self, dummy_spark):
        # Arrange
        te = TableExists(**self.options)
        expected_query = dedent(
            """
            SELECT *
            FROM INFORMATION_SCHEMA.TABLES
            WHERE TABLE_CATALOG     = 'DATABASE'
              AND TABLE_SCHEMA      = 'SCHEMA'
              AND TABLE_TYPE        = 'BASE TABLE'
              AND UPPER(TABLE_NAME) = 'TABLE'
            """
        ).strip()

        # Act
        output = te.execute()

        # Assert that the query is as expected and that we got exists as True
        assert dummy_spark.options_dict["query"] == expected_query
        assert output.exists


class TestTagSnowflakeQuery:
    def test_tag_query_no_existing_preactions(self):
        expected_preactions = (
            """ALTER SESSION SET QUERY_TAG = '{"pipeline_name": "test-pipeline-1","task_name": "test_task_1"}';"""
        )

        tagged_options = (
            TagSnowflakeQuery(
                task_name="test_task_1",
                pipeline_name="test-pipeline-1",
            )
            .execute()
            .options
        )

        assert len(tagged_options) == 1
        preactions = tagged_options["preactions"].replace("    ", "").replace("\n", "")
        assert preactions == expected_preactions

    def test_tag_query_present_existing_preactions(self):
        options = {
            "otherSfOption": "value",
            "preactions": "SET TEST_VAR = 'ABC';",
        }
        query_tag_preaction = (
            """ALTER SESSION SET QUERY_TAG = '{"pipeline_name": "test-pipeline-2","task_name": "test_task_2"}';"""
        )
        expected_preactions = f"SET TEST_VAR = 'ABC';{query_tag_preaction}"

        tagged_options = (
            TagSnowflakeQuery(task_name="test_task_2", pipeline_name="test-pipeline-2", options=options)
            .execute()
            .options
        )

        assert len(tagged_options) == 2
        assert tagged_options["otherSfOption"] == "value"
        preactions = tagged_options["preactions"].replace("    ", "").replace("\n", "")
        assert preactions == expected_preactions
